/*
 *	Sherlock Filter Engine
 *
 *	(c) 1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lib/lib.h"
#include "lib/pools.h"
#include "lib/url.h"
#include "filter.h"

struct filter *filter_current;

struct filter *
filter_load(char *name)
{
  struct filter *f;

  f = xmalloc(sizeof(struct filter));
  f->pool = new_pool(4096);

  filter_current = f;
  filter_lex_init(name);
  yyparse();
  filter_lex_cleanup();
  return f;
}

struct filter *
filter_load_default(void)
{
  return filter_load("cf/filter");
}

static uns
filter_fetch_num(struct filter_args *a, uns var)
{
  switch (var)
    {
    case F_VAR_PORT:
      return a->port;
    case F_VAR_TAG:
      return a->tag;
    case F_VAR_WEIGHT:
      return a->weight;
    case F_VAR_MAXOBJ:
      return a->maxobj;
    case F_VAR_EX_REFRESH:
      return a->ex_refresh;
    case F_VAR_EX_EXPIRE:
      return a->ex_expire;
    case F_VAR_EX_EXPIRE_ERR:
      return a->ex_expire_err;
    default:
      die("filter: Internal error: unknown numeric variable code %x", var);
    }
}

static void
filter_set_num(struct filter_args *a, uns var, uns val)
{
  switch (var)
    {
    case F_VAR_TAG:
      if (val > 255)
	{
	  log(L_ERROR "filter: Tag %d out of range", val);
	  return;
	}
      a->tag = val;
      break;
    case F_VAR_WEIGHT:
      a->weight = val;
      break;
    case F_VAR_MAXOBJ:
      a->maxobj = val;
      break;
    case F_VAR_EX_REFRESH:
      a->ex_refresh = val;
      break;
    case F_VAR_EX_EXPIRE:
      a->ex_expire = val;
      break;
    case F_VAR_EX_EXPIRE_ERR:
      a->ex_expire_err = val;
      break;
    default:
      die("filter: Internal error: unknown numeric variable code %x", var);
    }
}

static char *
filter_fetch_str(struct filter_args *a, uns var)
{
  char *str;

  if (var & F_VF_NUMERIC)
    {
      sprintf(a->buf, "%d", filter_fetch_num(a, var));
      return a->buf;
    }
  if (var >= F_VAR_ATTR_FIRST && var <= F_VAR_ATTR_LAST)
    {
      if (a->get_attr)
	str = a->get_attr(a, var - F_VAR_ATTR_FIRST);
      else
	str = NULL;
    }
  else
    switch (var)
      {
      case F_VAR_URL:
	str = a->url;
	break;
      case F_VAR_HOST:
	str = a->host;
	break;
      case F_VAR_SOURCE:
	str = a->source;
	break;
      case F_VAR_PATH:
	str = a->path;
	break;
      default:
	die("filter: Internal error: unknown string variable code %x", var);
      }
  return str ? str : "";
}

static int
filter_eval_cond(struct filter_cond *c, struct filter_args *a)
{
  switch (c->op)
    {
    case F_C_AND:
      return filter_eval_cond(c->u.expr.l, a) && filter_eval_cond(c->u.expr.r, a);
    case F_C_OR:
      return filter_eval_cond(c->u.expr.l, a) || filter_eval_cond(c->u.expr.r, a);
    case F_C_NOT:
      return !filter_eval_cond(c->u.expr.l, a);
    case F_C_MATCH_REGEX:
      return rx_match((regex *) c->u.match.patt, filter_fetch_str(a, c->u.match.what));
    case F_C_MATCH_PATTERN:
      return match_pattern(c->u.match.patt, filter_fetch_str(a, c->u.match.what));
    case F_C_MATCH_PATTERN_NOCASE:
      return match_pattern_nocase(c->u.match.patt, filter_fetch_str(a, c->u.match.what));
    case F_C_MATCH_EQUAL:
      return !strcmp(c->u.match.patt, filter_fetch_str(a, c->u.match.what));
    case F_C_MATCH_NUMBER_EQ:
      return (filter_fetch_num(a, c->u.cmp.what) == c->u.cmp.num);
    case F_C_MATCH_NUMBER_NE:
      return (filter_fetch_num(a, c->u.cmp.what) != c->u.cmp.num);
    case F_C_MATCH_NUMBER_LT:
      return (filter_fetch_num(a, c->u.cmp.what) < c->u.cmp.num);
    case F_C_MATCH_NUMBER_GT:
      return (filter_fetch_num(a, c->u.cmp.what) > c->u.cmp.num);
    case F_C_MATCH_NUMBER_LE:
      return (filter_fetch_num(a, c->u.cmp.what) <= c->u.cmp.num);
    case F_C_MATCH_NUMBER_GE:
      return (filter_fetch_num(a, c->u.cmp.what) >= c->u.cmp.num);
    default:
      die("filter: Internal error: unknown condition code %d", c->op);
    }
}

static int
filter_eval(struct filter_node *f, struct filter_args *a)
{
  int res;

  while (f)
    {
      switch (f->op)
	{
	case F_OP_ACCEPT:
	  return 1;
	case F_OP_REJECT:
	  return 0;
	case F_OP_LOG:
	  log(L_DEBUG "filter: %s", f->u.msg);
	  break;
	case F_OP_IF:
	  if (filter_eval_cond(f->u.cond.condition, a))
	    res = filter_eval(f->u.cond.positive, a);
	  else
	    res = filter_eval(f->u.cond.negative, a);
	  if (res >= 0)
	    return res;
	  break;
	case F_OP_SET_NUM:
	  filter_set_num(a, f->u.setnum.var, f->u.setnum.val);
	  break;
	case F_OP_INC_NUM:
	  filter_set_num(a, f->u.setnum.var, f->u.setnum.val + filter_fetch_num(a, f->u.setnum.var));
	  break;
	default:
	  die("filter: Internal error: unknown command code %d", f->op);
	}
      f = f->next;
    }
  return -1;
}

int
filter_run(struct filter_args *a)
{
  struct url url;
  char buf1[MAX_URL_SIZE], buf2[MAX_URL_SIZE];
  int res;

  if (res = url_canon_split(a->url, buf1, buf2, &url))
    {
      log(L_ERROR "filter: Error while parsing URL %s: %s", a->url, url_error(res));
      return 0;
    }
  a->host = url.host;
  a->port = url.port;
  a->path = url.rest;
  res = filter_eval(a->filter->body, a);
  if (res < 0)
    {
      log(L_ERROR "filter: Neither ACCEPT nor REJECT issued, defaulting to REJECT.");
      res = 0;
    }
  return res;
}
