/*
 *	Sherlock Filter Engine -- Lexer
 *
 *	(c) 1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

%{
#undef REJECT     /* Avoid name clashes */

#include "lib/lib.h"
#include "filter.h"
#include "parse.tab.h"

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#define err(x) filter_err(x)

#define KW(x) { #x, x }
static struct keyword {
  byte *name;
  int value;
} keyword_list[] = {
  KW(HOST),
  KW(SOURCE),
  KW(URL),
  KW(LOG),
  KW(ACCEPT),
  KW(REJECT),
  KW(IF),
  KW(ELSE),
  KW(ELSIF),
  KW(WEIGHT),
  KW(TAG),
  KW(PORT),
  KW(ATTR),
  KW(WEIGHT),
  KW(MAXOBJ),
  KW(PATH),
  KW(REFRESH_TIME),
  KW(EXPIRE_TIME),
  KW(ERR_EXPIRE_TIME),
  { NULL, -1 } };
#undef KW

static int lino;
static char *filter_name;

#define YY_FATAL_ERROR(msg) err(msg)

%}

%option noyywrap

%x COMMENT CCOMM

ALPHA [a-zA-Z_]
DIGIT [0-9]
XIGIT [0-9a-fA-F]
ALNUM [a-zA-Z_0-9]
WHITE [ \t]

%%

0x{DIGIT}+ {
  char *e;
  long int l;
  errno = 0;
  l = strtoul(yytext+2, &e, 16);
  if (e && *e || errno == ERANGE || (long int)(int) l != l)
    err("Number out of range");
  yylval.i = l;
  return NUM;
}

-?{DIGIT}+ {
  char *e;
  long int l;
  errno = 0;
  l = strtoul(yytext, &e, 10);
  if (e && *e || errno == ERANGE || (long int)(int) l != l)
    err("Number out of range");
  yylval.i = l;
  return NUM;
}

{ALPHA}{ALNUM}* {
  struct keyword *k = keyword_list;
  while (k->name)
    {
      if (!strcasecmp(k->name, yytext))
	return k->value;
      k++;
    }
  err("Unknown keyword");
}

\<=	return LE;
\>=	return GE;
@@	return ATAT;

[;&|!=~@(){}\[\]+<>-] {
  return yytext[0];
}

["][^"\n]*["] {
  yylval.s = yytext+1;
  yytext[yyleng-1] = 0;
  return STRING;
}

["][^"\n]*\n	err("Unterminated string");

<INITIAL,COMMENT><<EOF>>	return END;

{WHITE}+

\n	lino++;

#	BEGIN(COMMENT);

\/\*	BEGIN(CCOMM);

.	err("Unknown character");

<COMMENT>\n {
  lino++;
  BEGIN(INITIAL);
}

<COMMENT>.

<CCOMM>\*\/	BEGIN(INITIAL);
<CCOMM>\n	lino++;
<CCOMM>\/\*	err("Comment nesting not supported");
<CCOMM><<EOF>>	err("Unterminated comment");
<CCOMM>.

%%

void
filter_lex_init(char *name)
{
  lino = 1;
  filter_name = name;
  yyin = fopen(name, "r");
  if (!yyin)
    die("Unable to open %s: %m", name);
}

void
filter_lex_cleanup(void)
{
  fclose(yyin);
}

void
filter_err(char *msg)
{
  die("%s, line %d: %s", filter_name, lino, msg);
}
