/*
 *	Sherlock Filter Engine -- Parser
 *
 *	(c) 1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lib/lib.h"
#include "lib/pools.h"
#include "filter.h"

#define yyerror(x) filter_err(x)
#define err(x) filter_err(x)

void *filter_alloc(int size);
char *filter_copystr(char *str);
%}

%union {
  int i;
  char *s;
  struct filter_cond *c;
  struct filter_node *n;
}

/* Tokens */

%token END URL HOST SOURCE LOG ACCEPT REJECT IF ELSE WEIGHT TAG PORT ATTR
%token LE GE ATAT MAXOBJ ELSIF PATH REFRESH_TIME EXPIRE_TIME ERR_EXPIRE_TIME

%token <i> NUM
%token <s> STRING

%left '|'
%left '&'
%right '!'

/* Types */

%type <i> attr matchop maybenot simple_command matchopnum
%type <c> cond
%type <n> command block commands maybeelse

%%

main:
   commands END {
     filter_current->body = $1;
     return 0;
   }
 ;

cond:
   attr maybenot matchop STRING {
     $$ = filter_alloc(sizeof(struct filter_cond));
     $$->op = $3;
     $$->u.match.what = $1;
     if ($3 == F_C_MATCH_REGEX)
       {
	 regex *r = rx_compile($4);
	 if (!r)
	   err("Malformed regular expression");
	 $$->u.match.patt = (void *) r;
       }
     else
       $$->u.match.patt = filter_copystr($4);
     if ($2)
       {
	 struct filter_cond *n = $$;
	 $$ = filter_alloc(sizeof(struct filter_cond));
	 $$->op = F_C_NOT;
	 $$->u.expr.l = n;
	 $$->u.expr.r = NULL;
       }
   }
 | attr maybenot matchopnum NUM {
     $$ = filter_alloc(sizeof(struct filter_cond));
     $$->op = $3 ^ $2;
     $$->u.cmp.what = $1;
     $$->u.cmp.num = $4;
   }
 | cond '&' cond {
     $$ = filter_alloc(sizeof(struct filter_cond));
     $$->op = F_C_AND;
     $$->u.expr.l = $1;
     $$->u.expr.r = $3;
   }
 | cond '|' cond {
     $$ = filter_alloc(sizeof(struct filter_cond));
     $$->op = F_C_OR;
     $$->u.expr.l = $1;
     $$->u.expr.r = $3;
   }
 | '!' cond {
     $$ = filter_alloc(sizeof(struct filter_cond));
     $$->op = F_C_NOT;
     $$->u.expr.l = $2;
     $$->u.expr.r = NULL;
   }
 | '(' cond ')' { $$ = $2; }
 ;

attr:
   URL { $$ = F_VAR_URL; }
 | HOST { $$ = F_VAR_HOST; }
 | SOURCE { $$ = F_VAR_SOURCE; }
 | PORT { $$ = F_VAR_PORT; }
 | TAG { $$ = F_VAR_TAG; }
 | WEIGHT { $$ = F_VAR_WEIGHT; }
 | MAXOBJ { $$ = F_VAR_MAXOBJ; }
 | PATH { $$ = F_VAR_PATH; }
 | REFRESH_TIME { $$ = F_VAR_EX_REFRESH; }
 | EXPIRE_TIME { $$ = F_VAR_EX_EXPIRE; }
 | ERR_EXPIRE_TIME { $$ = F_VAR_EX_EXPIRE_ERR; }
 | ATTR '[' STRING ']' {
     if (strlen($3) != 1)
       err("Invalid attribute name");
     $$ = F_VAR_ATTR_FIRST + $3[0];
   }
 ;

matchop:
   '~' { $$ = F_C_MATCH_REGEX; }
 | '@' { $$ = F_C_MATCH_PATTERN; }
 | ATAT { $$ = F_C_MATCH_PATTERN_NOCASE; }
 | '=' { $$ = F_C_MATCH_EQUAL; }
 ;

matchopnum:
   '=' { $$ = F_C_MATCH_NUMBER_EQ; }
 | '<' { $$ = F_C_MATCH_NUMBER_LT; }
 | '>' { $$ = F_C_MATCH_NUMBER_GT; }
 | LE { $$ = F_C_MATCH_NUMBER_LE; }
 | GE { $$ = F_C_MATCH_NUMBER_GE; }
 ;

maybenot:
   /* empty */ { $$ = 0; }
 | '!' { $$ = 1; }
 ;

command:
   simple_command ';' {
     $$ = filter_alloc(sizeof(struct filter_node));
     $$->op = $1;
   }
 | LOG STRING ';' {
     $$ = filter_alloc(sizeof(struct filter_node));
     $$->op = F_OP_LOG;
     $$->u.msg = filter_copystr($2);
   }
 | IF cond block maybeelse {
     $$ = filter_alloc(sizeof(struct filter_node));
     $$->op = F_OP_IF;
     $$->u.cond.condition = $2;
     $$->u.cond.positive = $3;
     $$->u.cond.negative = $4;
   }
 | attr '=' NUM ';' {
     if (!($1 & F_VF_NUMERIC) || ($1 & F_VF_READONLY))
       err("This attribute cannot be set");
     $$ = filter_alloc(sizeof(struct filter_node));
     $$->op = F_OP_SET_NUM;
     $$->u.setnum.var = $1;
     $$->u.setnum.val = $3;
   }
 | attr '+' '=' NUM ';' {
     if (!($1 & F_VF_NUMERIC) || ($1 & F_VF_READONLY))
       err("This attribute cannot be set");
     $$ = filter_alloc(sizeof(struct filter_node));
     $$->op = F_OP_INC_NUM;
     $$->u.setnum.var = $1;
     $$->u.setnum.val = $4;
   }
 | attr '-' '=' NUM ';' {
     if (!($1 & F_VF_NUMERIC) || ($1 & F_VF_READONLY))
       err("This attribute cannot be set");
     $$ = filter_alloc(sizeof(struct filter_node));
     $$->op = F_OP_INC_NUM;
     $$->u.setnum.var = $1;
     $$->u.setnum.val = -$4;
   }
 ;

simple_command:
   ACCEPT { $$ = F_OP_ACCEPT; }
 | REJECT { $$ = F_OP_REJECT; }
 ;

maybeelse:
   /* empty */ { $$ = NULL; }
 | ELSE block { $$ = $2; }
 | ELSIF cond block maybeelse {
     $$ = filter_alloc(sizeof(struct filter_node));
     $$->op = F_OP_IF;
     $$->u.cond.condition = $2;
     $$->u.cond.positive = $3;
     $$->u.cond.negative = $4;
   }
 ;

block:
   '{' commands '}' { $$=$2; }
 ;

commands:
   command { $$=$1; $$->next=NULL; $$->last=$1; }
 | commands command {
     $$ = $1;
     $$->last->next = $2;
     $$->last = $2;
   }
 ;

%%

void *
filter_alloc(int size)
{
  return pool_alloc(filter_current->pool, size);
}

char *
filter_copystr(char *str)
{
  int len = strlen(str) + 1;
  char *x = filter_alloc(len);
  memcpy(x, str, len);
  return x;
}
