/*
 *	Sherlock Gatherer -- Data Type Identification
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>

#include "gather.h"

#include <lib/string.h>
#include <lib/url.h>
#include <lib/md5.h>

/* Structures */

struct intype_filter {			/* Input content type / content encoding filter */
  node n;
  byte *src, *dest;
};

struct crystal_ball {			/* Magic rules for datomanthy */
  node n;
  byte mode;				/* Crystal ball type */
  byte *rule;				/* Name or contents pattern */
  byte *type;				/* Resulting type */
};

#define CBMODE_NAME 0
#define CBMODE_BINARY 1
#define CBMODE_IS_ASCII 2

struct dt_to_enc {			/* Data type => encoding conversion */
  node n;
  byte *type;
  byte *encoding;
};

struct encoding {			/* Encoding */
  node n;
  byte *encoding;
  byte *drop_suffix;
  int enctype;
};

struct meta_type {			/* <META>s */
  node n;
  byte *type;
};

struct needing_charset {		/* Types needing charset detection and normalization */
  node n;				/* Also used for need_ascii */
  byte *type;
};

static byte *builtin_encodings[] = { "gzip", "deflate", NULL };

/* Configuration */

int trace_data_types;
#define TRACE (trace_data_types)
#define XTRACE (TRACE > 1)

static int max_convert;

static list intype_filters, crystal_balls, dt_to_encs, encodings, having_metas,
  inenc_filters, needing_charsets, needing_ascii;

static byte *snerr = "Syntax error";

static byte *
parse_intype(struct cfitem *c, byte *l)
{
  byte *w[2];
  struct intype_filter *f = xmalloc(sizeof(struct intype_filter));

  if (wordsplit(l, w, 2) != 2)
    return snerr;
  f->src = stralloc(w[0]);
  f->dest = stralloc(w[1]);
  add_tail(&intype_filters, &f->n);
  return NULL;
}

static byte *
parse_inenc(struct cfitem *c, byte *l)
{
  byte *w[2];
  struct intype_filter *f = xmalloc(sizeof(struct intype_filter));

  if (wordsplit(l, w, 2) != 2)
    return snerr;
  f->src = stralloc(w[0]);
  f->dest = stralloc(w[1]);
  add_tail(&inenc_filters, &f->n);
  return NULL;
}

static byte *
parse_isascii(struct cfitem *c, byte *l)
{
  struct crystal_ball *f = xmalloc(sizeof(struct crystal_ball));

  f->mode = CBMODE_IS_ASCII;
  f->rule = NULL;
  f->type = stralloc(l);
  add_tail(&crystal_balls, &f->n);
  return NULL;
}

static byte *
parse_filter(struct cfitem *c, byte *l)
{
  byte *w[2];
  struct crystal_ball *f = xmalloc(sizeof(struct crystal_ball));

  if (wordsplit(l, w, 2) != 2)
    return snerr;
  f->mode = (c->name[0] == 'B') ? CBMODE_BINARY : CBMODE_NAME;
  if (f->mode == CBMODE_BINARY)
    {
      int z, b = 0;
      byte *c = w[0];
      byte *d;

      while (*c)			/* Estimate length of compiled pattern */
	{
	  if (!Cdigit(*c))
	    return snerr;
	  while (Cdigit(*c))
	    c++;
	  if (*c++ != ':')
	    return snerr;
	  if (*c == '"')
	    {
	      c++;
	      while (*c != '"')
		{
		  if (!*c)
		    return snerr;
		  c++, b += 2;
		}
	      c++;
	    }
	  else if (Cxdigit(*c))
	    {
	      while (Cxdigit(*c))
		c++, b++;
	      if (b & 1)
		return snerr;
	    }
	  else
	    return snerr;
	  if (*c)
	    {
	      if (*c == ',')
		c++;
	      else
		return snerr;
	    }
	}

      d = f->rule = xmalloc(b + 1);	/* Compile the pattern */
      c = w[0];
      while (*c)
	{
	  b = 0;
	  while (*c != ':')
	    b = 10*b + *c++ - '0';
	  if (b < 0 || b > 254)
	    return "Position out of range";
	  if (*++c == '"')
	    {
	      c++;
	      while (*c != '"')
		{
		  *d++ = b++;
		  *d++ = *c++;
		}
	      c++;
	    }
	  else while (Cxdigit(*c))
	    {
	      int d0, d1;
	      *d++ = b++;
	      d0 = *c++;
	      d1 = *c++;
	      z = (Cxvalue(d0) << 4) | Cxvalue(d1);
	      *d++ = z;
	    }
	  if (b > 254)
	    return "Position out of range";
	  if (*c)
	    c++;
	}
      *d = 0xff;
    }
  else
    f->rule = stralloc(w[0]);
  f->type = stralloc(w[1]);
  add_tail(&crystal_balls, &f->n);
  return NULL;
}

static byte *
parse_dtenc(struct cfitem *c, byte *l)
{
  byte *w[2];
  struct dt_to_enc *f = xmalloc(sizeof(struct dt_to_enc));

  if (wordsplit(l, w, 2) != 2)
    return snerr;
  f->type = stralloc(w[0]);
  f->encoding = stralloc(w[1]);
  add_tail(&dt_to_encs, &f->n);
  return NULL;
}

static byte *
parse_encoding(struct cfitem *c, byte *l)
{
  byte *w[2];
  struct encoding *f = xmalloc(sizeof(struct encoding));
  int z, e;

  z = wordsplit(l, w, 2);
  if (z != 1 && z != 2)
    return snerr;
  for(e=0; builtin_encodings[e]; e++)
    if (!strcmp(builtin_encodings[e], w[0]))
      {
	f->enctype = -e-1;
	f->encoding = stralloc(w[0]);
	f->drop_suffix = (z == 2) ? stralloc(w[1]) : NULL;
	add_tail(&encodings, &f->n);
	return NULL;
      }
  return "Unknown content encoding";
}

static byte *
parse_dometas(struct cfitem *c, byte *l)
{
  struct meta_type *f = xmalloc(sizeof(struct meta_type));

  f->type = stralloc(l);
  add_tail(&having_metas, &f->n);
  return NULL;
}

static byte *
parse_need_charset(struct cfitem *c, byte *l)
{
  struct needing_charset *f = xmalloc(sizeof(struct needing_charset));

  f->type = stralloc(l);
  add_tail(&needing_charsets, &f->n);
  return NULL;
}

static byte *
parse_need_ascii(struct cfitem *c, byte *l)
{
  struct needing_charset *f = xmalloc(sizeof(struct needing_charset));

  f->type = stralloc(l);
  add_tail(&needing_ascii, &f->n);
  return NULL;
}

static struct cfitem cfitems[] = {
  { "Trace", CI_INT, &trace_data_types },
  { "MaxConvert", CI_INT, &max_convert },
  { "InType", CI_FUNCTION, parse_intype },
  { "InEnc", CI_FUNCTION, parse_inenc },
  { "Name", CI_FUNCTION, parse_filter },
  { "Bytes", CI_FUNCTION, parse_filter },
  { "IsAscii", CI_FUNCTION, parse_isascii },
  { "DTEnc", CI_FUNCTION, parse_dtenc },
  { "Encoding", CI_FUNCTION, parse_encoding },
  { "MaxDecompress", CI_INT, &max_decode },
  { "DoMetas", CI_FUNCTION, parse_dometas },
  { "NeedCharset", CI_FUNCTION, parse_need_charset },
  { "NeedAscii", CI_FUNCTION, parse_need_ascii },
  { NULL, CI_STOP, NULL }
};

void
datatype_init(void)
{
  init_list(&intype_filters);
  init_list(&inenc_filters);
  init_list(&crystal_balls);
  init_list(&dt_to_encs);
  init_list(&encodings);
  init_list(&having_metas);
  init_list(&needing_charsets);
  init_list(&needing_ascii);
  cf_read("cf/datatypes", cfitems);
}

/* Information extracted from current file */

static byte *url, *base, *host, *data_file;
static byte datatype[64];
static byte encoding[64];
static byte charset[64];
static byte url1[MAX_URL_SIZE], url2[MAX_URL_SIZE], basebuf[MAX_URL_SIZE];

static void				/* Extract URL, datatype, encoding and charset from the object description */
extract_odes(struct odes *o)
{
  byte *b, *c, *d, *e;
  int err;
  struct url u;

  if (! (url = find_aval(o, 'U')))
    url = "<UNKNOWN>";
  if (XTRACE)
    log(L_DEBUG "Data type analysis for %s", url);

  datatype[0] = 0;
  encoding[0] = 0;
  charset[0] = 0;

  if (b = find_aval(o, 'T'))
    {
      if (d = strchr(b, ';'))
	{
	  c = d;
	  *d++ = 0;
	}
      else
	c = b + strlen(b);
      while (c > b && Cspace(c[-1]))
	*--c = 0;
      if (strlen(b) > sizeof(datatype) - 1)
	log(L_ERROR "%s: Content-type spec too long", url);
      else
	{
	  strcpy(datatype, b);
	  while (d)
	    {
	      if (!*d)
		break;
	      if (Cspace(*d))
		{
		  d++;
		  continue;
		}
	      if (e = strchr(d, '='))
		{
		  *e++ = 0;		/* d=parameter, e=value */
		  c = e;
		  if (*c == '"')	/* Quoted value */
		    {
		      c++, e++;
		      while (*c && *c != '"')
			c++;
		      *c++ = 0;
		      c = strchr(c, ';');
		    }
		  else			/* Unquoted value => skip trailing spaces */
		    {
		      b = c = strchr(c, ';');
		      if (!b)
			b = e + strlen(e);
		      while (b > e && Cspace(b[-1]))
			*--b = 0;
		    }
		  if (c)
		    *c++ = 0;
		  if (strlen(e) > sizeof(charset) - 1)
		    log(L_ERROR "%s: Content-type extension parameter too long", url);
		  else if (!strcasecmp(d, "charset"))
		    strcpy(charset, e);
		  else if (TRACE)
		    log(L_DEBUG "Unknown content-type extension: %s=%s", d, e);
		}
	      else
		c = NULL;
	      d = c;
	    }
	}
    }

  if (b = find_aval(o, 'c'))
    {
      if (strlen(b) > sizeof(charset) - 1)
	log(L_ERROR "%s: Charset field too long", url);
      else
	strcpy(charset, b);
    }

  if (b = find_aval(o, 'E'))
    {
      if (strlen(b) > sizeof(encoding) - 1)
	log(L_ERROR "%s: Encoding spec too long", url);
      else
	strcpy(encoding, b);
    }

  if ((err = url_deescape(url, url1)) || (err = url_split(url1, &u, url2)))
    die("%s: URL parse error %d in extract_odes", url, err);
  if (! (host = u.host))
    host = "UNKNOWN";
  if (b = find_aval(o, 'N'))
    {
      strcpy(basebuf, b);
      base = basebuf;
    }
  else
    {
      url_enescape(u.rest, basebuf);
      if (c = strrchr(basebuf, '/'))
	base = c + 1;
      else
	base = basebuf;
    }

  if (TRACE)
    log(L_DEBUG "CT: type=`%s', charset=`%s', encoding=`%s', base=`%s'", datatype, charset, encoding, base);

  if (datatype[0] && (! (b = strchr(datatype, '/')) || strchr(b+1, '/')))
    {
      log(L_ERROR "Invalid Content-type `%s' got for %s", datatype, url);
      datatype[0] = 0;
    }
}

static int
translate_type(void)			/* Perform input filtering of data types */
{
  struct intype_filter *f;

  DO_FOR_ALL(f, intype_filters)
    if (match_ct_patt(f->src, datatype))
      {
	if (TRACE)
	  log(L_DEBUG "In filter: type=`%s'", f->dest);
	if (!strcmp(f->dest, "ERROR"))
	  {
	    log(L_ERROR "%s: Input filter reports suspicious data type (%s)", url, datatype);
	    return 1;
	  }
	else if (!strcmp(f->dest, "UNKNOWN"))
	  datatype[0] = 0;
	else
	  strcpy(datatype, f->dest);
	return 0;
      }
  return 0;
}

static int
translate_encoding(void)		/* Perform input filtering of content encodings */
{
  struct intype_filter *f;

  DO_FOR_ALL(f, inenc_filters)
    if (match_pattern(f->src, encoding))
      {
	if (TRACE)
	  log(L_DEBUG "In filter: encoding=`%s'", f->dest);
	if (!strcmp(f->dest, "ERROR"))
	  {
	    log(L_ERROR "%s: Input filter reports suspicious content encoding (%s)", url, encoding);
	    return 1;
	  }
	else if (!strcmp(f->dest, "UNKNOWN"))
	  encoding[0] = 0;
	else
	  strcpy(encoding, f->dest);
	return 0;
      }
  return 0;
}

/* Data type vaticination */

int dt_fh;
uns dt_in_size, dt_where, dt_last;
byte dt_inbuf[1025];			/* The post-last byte is _always_ zero */

static int				/* Open the data file */
dt_open(byte *name)
{
  if (!name)
    {
      log(L_ERROR "dt_open: Name unknown");
      return -1;
    }
  data_file = name;
  dt_where = ~0;
  if ((dt_fh = open(name, O_RDONLY, 0)) < 0)
    {
      log(L_ERROR "dt_open(%s): %m", name);
      return -1;
    }
  return 0;
}

int					/* Read next block */
dt_read_next(void)
{
  int s;

  if (dt_fh < 0)
    return dt_fh;
  if ((s = read(dt_fh, dt_inbuf, sizeof(dt_inbuf) - 1)) < 0)
    {
      log(L_ERROR "%s: Read error on `%s': %m", url, data_file);
      close(dt_fh);
      return dt_fh = -1;
    }
  dt_in_size = s;
  dt_inbuf[dt_in_size] = 0;
  dt_where = dt_last;
  dt_last += dt_in_size;
  return dt_in_size;
}

int					/* Read first block */
dt_read_head(void)
{
  if (dt_fh < 0)
    return dt_fh;
  if (!dt_where)
    return 0;
  if (dt_where != ~0U)
    lseek(dt_fh, 0, SEEK_SET);
  dt_where = dt_last = 0;
  return dt_read_next();
}

static int
is_ascii(void)
{
  byte *k;
  uns l;

  if (dt_read_head() < 0)
    return 0;
  for(k=dt_inbuf, l=dt_in_size; l; k++, l--)
    if (*k < 0x20 && *k != 0x0d && *k != 0x0a && *k != 0x0c && *k != 0x09
	|| *k == 0x7f)
      return 0;
  return 1;
}

static void
vaticinate(void)			/* Go to local oracle and ask Pythia about the data type */
{
  struct crystal_ball *f;
  int i = 0;

  DO_FOR_ALL(f, crystal_balls)
    {
      i++;
      switch (f->mode)
	{
	case CBMODE_BINARY:
	  {
	    byte *r = f->rule;
	    uns k, l;
	    
	    if (dt_read_head() < 0)
	      continue;
	    while (*r != 0xff)
	      {
		k = *r++;		/* Address */
		l = *r++;		/* Data */
		if (k >= dt_in_size || l != dt_inbuf[k])
		  goto cont;
	      }
	    break;
	  }
	case CBMODE_NAME:
	  if (!match_pattern_nocase(f->rule, base))
	    continue;
	  break;
	case CBMODE_IS_ASCII:
	  if (!is_ascii())
	    continue;
	  break;
	default:
	  die("Unknown crystal ball #%d", f->mode);
	}
      if (TRACE)
	log(L_DEBUG "Rule #%d: %s", i, f->type);
      if (!strcmp(f->type, "UNKNOWN"))
	datatype[0] = 0;
      else
	strcpy(datatype, f->type);
      return;
    cont:
    }
}

static void
maybe_encoding(void)			/* Maybe our data type is just an encoding? */
{
  struct dt_to_enc *f;

  DO_FOR_ALL(f, dt_to_encs)
    if (match_ct_patt(f->type, datatype))
      {
	if (TRACE)
	  log(L_DEBUG "Data type %s -> Encoding %s", datatype, f->encoding);
	datatype[0] = 0;
	strcpy(encoding, f->encoding);
	return;
      }
}

/* Content encodings */

static inline void
chop_off_suffix(byte *s)		/* Chop off filename suffix when decoding */
{
  int l = strlen(base);
  int k = strlen(s);

  if (l >= k && !strcmp(base + l - k, s))
    {
      base[l - k] = 0;
      if (TRACE)
	log(L_DEBUG "Renaming to `%s'", base);
    }
}

static int
which_decoder(void)			/* Which decoder should we use? */
{
  struct encoding *e;

  translate_encoding();
  if (!encoding[0])
    return 0;

  DO_FOR_ALL(e, encodings)
    if (!strcasecmp(e->encoding, encoding))
      {
	if (TRACE)
	  log(L_DEBUG "Will decode by algorithm %d", e->enctype);
	if (e->drop_suffix)
	  chop_off_suffix(e->drop_suffix);
	return e->enctype;
      }
  return 0;
}

/* <META> */

static int
should_meta(void)
{
  struct meta_type *f;

  DO_FOR_ALL(f, having_metas)
    if (match_ct_patt(f->type, datatype))
      return 1;
  return 0;
}

static int
scan_meta(struct odes *o)
{
  byte metabuf[256];

  if (XTRACE)
    log(L_DEBUG "Scanning for hidden META Content-Type");

  if (dt_read_head() < 0)
    return 0;
  if (hunt_meta(dt_inbuf, metabuf))
    {
      if (TRACE)
	log(L_INFO "%s: META Content-Type %s", url, metabuf);
      set_attr(o, 'T', metabuf);
      add_attr(o, NULL, '.', "META Content-Type found");
      set_attr(o, 'c', NULL);
      return 1;
    }
  return 0;
}

/* Transformation and checksumming of file data */

static void
transform_data(struct odes *o, int recode)
{
  int df;
  struct tempfile tf;
  struct MD5Context md;
  byte digest[MD5_SIZE], hexdigest[MD5_HEX_SIZE];
  byte *dest = NULL, *dp = NULL;
  uns ds = 1024;
  uns dbsize = 1024;

  if (recode)				/* Need to recode? */
    {
      open_temp(&tf, TF_TRANSFORM);
      prepend_attr(o, '@', tf.name);
      df = tf.fh;
      dest = dp = alloca(dbsize);
      ds = dbsize;
      char_convert_init(o);
    }
  else
    df = -1;

  MD5Init(&md);
  if (dt_read_head() < 0)
    goto done;
  while (dt_in_size)
    {
      byte *w = dt_inbuf;
      uns pos = dt_where;
      uns end = dt_last;
      uns howmuch;
      while (pos < end)
	{
	  if (meta_end && pos >= meta_start && pos < meta_end)
	    {
	      uns sk = meta_end - pos;
	      w += sk;
	      pos += sk;
	      continue;
	    }
	  if (meta_end && pos < meta_start && end > meta_start)
	    howmuch = meta_start - pos;
	  else
	    howmuch = end - pos;
	  if (df >= 0)			/* Recoding */
	    {
	      uns rem = howmuch;
	      byte *ww = w;
	      char_convert(&ww, &rem, &dp, &ds);
	      if (!ds)
		{
		  uns s = dp - dest;
		  if ((uns) write(df, dest, s) != s)
		    log(L_ERROR "Write error on temp file: %m");
		  MD5Update(&md, dest, s);
		  dp = dest;
		  ds = dbsize;
		}
	      howmuch = ww - w;
	    }
	  else
	    MD5Update(&md, w, howmuch);
	  pos += howmuch;
	  w += howmuch;
	}
      if (dt_read_next() < 0)
	goto done;
    }
done:
  if (df >= 0 && dp != dest)
    {
      uns s = dp - dest;
      if ((uns) write(df, dest, s) != s)
	log(L_ERROR "Write error on temp file: %m");
      MD5Update(&md, dest, s);
    }
  MD5Final(digest, &md);
  md5_to_hex(digest, hexdigest);
  set_attr(o, 'M', hexdigest);
  if (df >= 0)
    close(df);
}

/* Charsets */

static int
needs_charset(void)
{
  struct needing_charset *x;

  DO_FOR_ALL(x, needing_charsets)
    if (match_ct_patt(x->type, datatype))
      return 1;
  return 0;
}

static int
needs_ascii(void)
{
  struct needing_charset *x;

  DO_FOR_ALL(x, needing_ascii)
    if (match_ct_patt(x->type, datatype))
      return 1;
  return 0;
}

/* DATA TYPE PROCESSING -- MAIN ENTRY POINT */

int
datatype_guess(struct odes *o)
{
  int rc = 0;
  int ce;
  int ttl = 0;
  int did_meta = 0;
  byte *b;

  meta_start = meta_end = 0;		/* No <META ...> information found */
  if (dt_open(find_aval(o, '@')))	/* Open the data file */
    return 404;

again:
  extract_odes(o);			/* Extract object data */

  if (b = find_aval(o, 'x'))		/* Check file TTL */
    ttl = atol(b);
  if (ttl >= max_convert)
    {
      log(L_ERROR "%s: Too many type conversions (maybe an infinite loop?)", url);
      return 402;
    }

  if (datatype[0] && translate_type())	/* Determine correct data type */
    return 401;
  if (datatype[0] && needs_ascii() && !is_ascii())
    datatype[0] = 0;
  if (!datatype[0])
    vaticinate();
  if (datatype[0])
    maybe_encoding();

  if (ce = which_decoder())		/* Should we decode? [suffix chopping done here as well] */
    {
      set_attr(o, 'n', encoding);
      encoding[0] = 0;
      rc = ce;
    }
  else if (encoding[0])			/* Unknown encoding => reject */
    {
      log(L_ERROR "%s: Unsupported content encoding `%s'", url, encoding);
      rc = 403;
    }
  else
    {
      encoding[0] = 0;
      if (!datatype[0])			/* Unknown data type and guessing failed :-( */
	{
	  if (TRACE)
	    log(L_DEBUG "Cannot determine data type");
	  rc = 400;
	}
      if (!did_meta++ && should_meta() && scan_meta(o)) /* Scan for <META>? */
	goto again;
    }

  set_attr(o, 'T', datatype[0] ? datatype : NULL);
  set_attr(o, 'c', charset[0] ? charset : NULL);
  set_attr(o, 'E', encoding[0] ? encoding : NULL);
  set_attr(o, 'N', base);
  set_attr_num(o, 'x', ttl+1);

  if (!rc)
    {
      int recharset = 0;		/* Handle charsets */
      if (needs_charset())
	recharset = guess_charset(o, url, host);
      transform_data(o, recharset);	/* Transform file data and compute MD5 sum */
    }

  close(dt_fh);
  return rc;
}
