/*
 *	Sherlock Gatherer -- Databases
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>

#include "gather.h"
#include "lib/db.h"
#include "lib/url.h"

/** CONFIGURATION **/

static uns urldb_cache_size = 16;
static uns md5db_cache_size = 16;

static struct cfitem cfitems[] = {
  { "URLDbCacheSize", CI_INT, &urldb_cache_size },
  { "MD5DbCacheSize", CI_INT, &md5db_cache_size },
  { NULL, CI_STOP, NULL }
};

void
db_init(void)
{
  cf_read("cf/dbm", cfitems);
}

/** THE GATHERER LOCK ***/

void
gather_lock(void)
{
  int fd;

  fd = open(LOCK_NAME, O_WRONLY | O_CREAT | O_EXCL, 0666);
  if (fd < 0)
    die("Cannot get gatherer lock: %m");
}

void
gather_unlock(void)
{
  if (unlink(LOCK_NAME) < 0)
    die("Cannot free gatherer lock: %m");
}

/*** THE URL DATABASE ***/

static struct sdbm *urldb;

void
urldb_open(void)
{
  struct sdbm_options o;

  o.name = URLDB_NAME;
  o.flags = SDBM_CREAT | SDBM_WRITE;
  o.page_order = 11;	/* We need >=11 because or URL size and 11 gives the best performance */
  o.cache_size = urldb_cache_size;
  o.key_size = -1;
  o.val_size = sizeof(struct urlrec);
  if (!(urldb = sdbm_open(&o)))
    die("Cannot open the URL database");
}

void
urldb_close(void)
{
  sdbm_close(urldb);
}

int
urldb_lookup(byte *url, struct urlrec *u)
{
  uns urs = sizeof(struct urlrec);

  return sdbm_fetch(urldb, url, strlen(url), (byte *) u, &urs);
}

int
urldb_exists(byte *url)
{
  return sdbm_fetch(urldb, url, sizeof(url), NULL, NULL);
}

void
urldb_rewind(void)
{
  sdbm_rewind(urldb);
}

byte *
urldb_get_next(byte *buf, struct urlrec *u)
{
  uns keysize = MAX_URL_SIZE - 1;
  uns valsize = sizeof(struct urlrec);

  if (sdbm_get_next(urldb, buf, &keysize, (byte *) u, &valsize) > 0)
    {
      buf[keysize] = 0;
      return buf;
    }
  else
    return NULL;
}

int
urldb_delete(byte *url)
{
  return sdbm_delete(urldb, url, strlen(url));
}

void
urldb_store(byte *url, u32 stat, oid_t oid)
{
  struct urlrec ur = { stat, oid };

  sdbm_replace(urldb, url, strlen(url), (byte *) &ur, sizeof(ur));
}

/*** THE MD5 SUM DATABASE ***/

static struct sdbm *md5db;

void
md5db_open(void)
{
  struct sdbm_options o;

  o.name = MD5DB_NAME;
  o.flags = SDBM_CREAT | SDBM_WRITE;
  o.page_order = 10;
  o.cache_size = md5db_cache_size;
  o.key_size = MD5_SIZE;
  o.val_size = sizeof(struct md5rec);
  if (!(md5db = sdbm_open(&o)))
    die("Cannot open the MD5 database");
}

void
md5db_close(void)
{
  sdbm_close(md5db);
}

int
md5db_lookup(byte *md5, struct md5rec *m)
{
  uns ms = sizeof(struct md5rec);

  return sdbm_fetch(md5db, md5, MD5_SIZE, (byte *) m, &ms);
}

int
md5db_exists(byte *md5)
{
  return sdbm_fetch(md5db, md5, MD5_SIZE, NULL, NULL);
}

void
md5db_rewind(void)
{
  sdbm_rewind(md5db);
}

byte *
md5db_get_next(byte *key, struct md5rec *m)
{
  uns keysize = MD5_SIZE;
  uns valsize = sizeof(struct md5rec);

  if (sdbm_get_next(md5db, key, &keysize, (byte *) m, &valsize))
    return key;
  else
    return NULL;
}

int
md5db_delete(byte *md5)
{
  return sdbm_delete(md5db, md5, MD5_SIZE);
}

void
md5db_store(byte *md5, oid_t oid)
{
  struct md5rec md = { oid };

  sdbm_replace(md5db, md5, MD5_SIZE, (byte *) &md, sizeof(md));
}
