/*
 *	Sherlock Gatherer Database Expiration Utility
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <getopt.h>
#include <dirent.h>
#include <sys/stat.h>

#include "gather.h"

#include <lib/url.h>
#include <lib/fastbuf.h>
#include "filter/filter.h"

/*** CONFIGURATION ***/

static int verbose;
static byte *log_file;
static uns refresh_after = ~0, expire_after = ~0, expire_err_after = ~0;

uns show_forbidden;			/* Exported to url_filter.c */

static struct cfitem cfitems[] = {
  { "LogFile", CI_STRING, &log_file },
  { "RefreshAfter", CI_INT, &refresh_after },
  { "ExpireAfter", CI_INT, &expire_after },
  { "ExpireErrorsAfter", CI_INT, &expire_err_after },
  { "Verbose", CI_INT, &verbose },
  { NULL, CI_STOP, NULL }
};

/*** OTHER GLOBAL VARIABLES ***/

static int rem_empty, rebuild_queue;
static struct filter *extra_filter;

static uns obj_count, ignore_count, requeue_count, expire_count, filter_count;
static uns empty_dir_count;
static struct fastbuf *F;

/*** REBUILDING OF QUEUE ***/

static void
queue_rebuild_host(oid_t oid, byte *name)
{
  struct odes *o;
  struct qhost *h;

  o = obj_load(name);
  h = new_host(atoi(find_aval(o, 'p')),
	       find_aval(o, 'H'),
	       atoi(find_aval(o, 'P')),
	       oid);
  obj_free(o);
  enqueue_host(h, 1);
}

static void
queue_rebuild_start(void)
{
  log(L_INFO "Rebuilding queue");
  unlink("db/queue");
  unlink("db/hosts");
  queue_init();
  scan_obj_tree("host", queue_rebuild_host);
}

static void
queue_rebuild_object(struct urlrec *r, byte *url)
{
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE];
  struct url u;
  int err;

  if (err = url_canon_split(url, b1, b2, &u))
    log(L_ERROR "Error parsing %s: %d", url, err);
  else
    {
      struct qhost *h = find_host(u.protoid, u.host, u.port);
      if (!h)
	log(L_ERROR "No host for %s", url);
      else
	{
	  if (r->oid != OID_AUX)
	    h->obj_count++;
	  if (r->access & 1)
	    enqueue_item(h, u.rest);
	}
    }
}

/*** SCANNING OF OBJECTS ***/

static char *
try_filter(byte *url, struct filter_args *a)
{
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE], u0[MAX_URL_SIZE];
  struct url uu;
  int err;

  bzero(a, sizeof(*a));
  a->ex_refresh = refresh_after;
  a->ex_expire = expire_after;
  a->ex_expire_err = expire_err_after;
  strcpy(u0, url);
  if (err = url_canon_split(u0, b1, b2, &uu))
    {
      log(L_ERROR "Error parsing `%s': %s", u0, url_error(err));
      return "URL";
    }
  if (translate_url(&uu, u0, b1, b2) < 0)
    return "xlate";
  if (!okay_to_get(NULL, &uu, a))
    return "global";
  if (extra_filter)
    {
      byte url1[MAX_URL_SIZE], url[MAX_URL_SIZE];
      url_pack(&uu, url1);
      url_enescape(url1, url);
      a->filter = extra_filter;
      a->url = url;
      if (!filter_run(a))
	return "extra";
    }
  return NULL;
}

static void
expire_schedule(struct urlrec *r, byte *name, int *counter)
{
  if (!rebuild_queue && (r->access & 1))
    {
      /* Should be expired, but it's queued. Pity. */
      ignore_count++;
      return;
    }
  r->access &= ~1;
  (*counter)++;
  bputc(F, '-');
  bputsn(F, name);
}

static void
expire_fill(int do_expire)
{
  byte buf[MAX_URL_SIZE];
  struct urlrec r;
  uns now = time(NULL);
  uns age;
  struct filter_args a;
  char *msg;

  log("Scanning objects");
  urldb_rewind();
  while (urldb_get_next(buf, &r))
    {
      obj_count++;
      age = now - r.access;

      if (msg = try_filter(buf, &a))	/* Filtering out */
	{
	  if (verbose)
	    log(L_INFO "%s: Filtering out (%s, age=%d)", buf, msg, age);
	  if (do_expire)
	    expire_schedule(&r, buf, &filter_count);
	}
      else if (r.access > now)
	log(L_ERROR "%s: Ante-dated (access=%d, now=%d)", buf, r.access, now);
      else if (r.oid && r.oid < OID_MIN && r.oid != OID_AUX)
	{
	  if (do_expire && age >= a.ex_expire_err)
	    {
	      if (verbose)
		log(L_INFO "%s [%d]: Expiring (age=%d)", buf, r.oid, age);
	      expire_schedule(&r, buf, &expire_count);
	    }
	  else if (verbose > 1)
	    log(L_DEBUG "%s [%d]: OK", buf, r.oid);
	}
      else if (do_expire && age >= a.ex_expire)
	{
	  if (verbose)
	    log(L_INFO "%s: Expiring (age=%d)", buf, age);
	  expire_schedule(&r, buf, &expire_count);
	}
      else if (do_expire && age >= a.ex_refresh && !(r.access & 1))
	{
	  if (verbose)
	    log(L_INFO "%s: Requeued (age=%d)", buf, age);
	  requeue_count++;
	  bputc(F, '+');
	  bputsn(F, buf);
	}
      else if (verbose > 1)
	log(L_DEBUG "%s: OK (age=%d)", buf, age);

      if (rebuild_queue)
	queue_rebuild_object(&r, buf);
    }
}

/*** ACTUAL DELETION OF OBJECTS ***/

static void
requeue_obj(struct urlrec *r, byte *url)
{
  struct url u;
  int err;
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE];
  struct qhost *h;

  err = url_canon_split(url, b1, b2, &u);
  if (err)
    {
      log(L_ERROR "%s: %s", url, url_error(err));
      return;
    }
  h = find_host(u.protoid, u.host, u.port);
  if (!h)
    {
      log(L_ERROR "%s: Host %d:%s:%d not found", url, u.protoid, u.host, u.port);
      return;
    }
  enqueue_item(h, u.rest);
  enqueue_host(h, 1);
  urldb_store(url, r->access|1, r->oid);
}

static void
dec_objcount(struct urlrec *r, byte *url)
{
  struct url u;
  int err;
  byte b1[MAX_URL_SIZE], b2[MAX_URL_SIZE];
  struct qhost *h;

  if (r->oid == OID_AUX)
    return;
  err = url_canon_split(url, b1, b2, &u);
  if (err)
    {
      log(L_ERROR "%s: %s", url, url_error(err));
      return;
    }
  h = find_host(u.protoid, u.host, u.port);
  if (!h)
    {
      log(L_ERROR "%s: Host %d:%s:%d not found", url, u.protoid, u.host, u.port);
      return;
    }
  if (h->obj_count)
    h->obj_count--;
  else
    log(L_ERROR "obj_count found zero while deleting %s", url);
}

static void
kill_obj(struct urlrec *r, byte *url)
{
  byte name[64];
  struct odes *d;
  struct oattr *a;
  byte *x;

  if (r->oid < OID_MIN)
    goto delurl;
  mk_obj_name("obj", r->oid, name);
  d = obj_load(name);
  a = find_attr(d, 'U');
  while (a)
    {
      if (!strcmp(a->val, url))
	break;
      a = a->same;
    }
  if (!a)
    {
      log(L_ERROR "%s: object %08x doesn't contain this URL", url, r->oid);
    erex:
      obj_free(d);
      return;
    }
  if (!del_attr(d, a))
    {
      log(L_ERROR "%s: del_attr failed", url);
      goto erex;
    }
  if (find_aval(d, 'U'))
    {
      if (verbose)
	log(L_INFO "Object %08x has more references", r->oid);
      obj_write(name, d);
      obj_free(d);
      goto delurl;
    }
  if (x = find_aval(d, 'M'))
    {
      byte md5[MD5_SIZE];
      if (verbose > 5)
	log(L_DEBUG "Deleting MD5 %s", x);
      hex_to_md5(x, md5);
      if (!md5db_delete(md5))
	{
	  log(L_ERROR "Error while deleting MD5 %s for %s", x, url);
	  goto erex;
	}
    }
  obj_free(d);
  if (unlink(name) < 0)
    log(L_ERROR "unlink(%s) failed: %m", name);

delurl:
  if (!urldb_delete(url))
    log(L_ERROR "Error while deleting URL %s", url);
  dec_objcount(r, url);
}

static void
expire_read(void)
{
  byte buf[MAX_URL_SIZE+4];
  struct urlrec r;
  byte *url = buf+1;

  log(L_INFO "Deleting objects");
  while (bgets(F, buf, sizeof(buf)))
    {
      if (verbose > 3)
	puts(buf);
      if (!urldb_lookup(url, &r))
	die("%s: No data found", url);
      /*
       * In every case, we know the object is not queued, although it might
       * have the `queued' flag set.
       */
      switch (buf[0])
	{
	case '+':
	  requeue_obj(&r, url);
	  break;
	case '-':
	  kill_obj(&r, url);
	  rem_empty = 1;
	  break;
	default:
	  die("%s: Unknown operation");
	}
    }
}

/*** SCANNING OF DIRECTORIES ***/

static uns
do_scan_dirs(byte *name, byte *p, byte *end)
{
  DIR *d;
  int empty = 1;
  struct dirent *e;
  struct stat st;

  *p = 0;
  d = opendir(name);
  if (!d)
    {
      log(L_ERROR "Unable to scan %s: %m", name);
      return 0;
    }
  while (e = readdir(d))
    {
      byte *n = e->d_name;
      int l = strlen(n);

      if (p + l + 2 >= end)
	{
	  log(L_ERROR "Path too long");
	  continue;
	}
      strcpy(p, n);
      if (stat(name, &st) < 0)
	{
	  log(L_ERROR "stat %s: %m", name);
	  goto notempty;
	}
      if (S_ISDIR(st.st_mode))
	{
	  if (!strcmp(n, ".") || !strcmp(n, ".."))
	    continue;
	  p[l] = '/';
	  if (do_scan_dirs(name, p+l+1, end))
	    {
	      p[l] = 0;
	      if (lstat(name, &st) < 0)
		{
		  log(L_ERROR "lstat %s: %m", name);
		  goto notempty;
		}
	      if (S_ISLNK(st.st_mode))
		goto notempty;
	      if (verbose)
		log(L_INFO "Removing dir %s", name);
	      if (rmdir(name))
		log(L_ERROR "Unable to remove %s: %m", name);
	      else
		empty_dir_count++;
	      continue;
	    }
	}
    notempty:
	empty = 0;
    }
  closedir(d);
  return empty;
}

static void
scan_empty_dirs(byte *base)
{
  byte name[256];

  strcpy(name, base);
  do_scan_dirs(name, name + strlen(base), name + sizeof(name) - 2);
}

static void
rem_empty_dirs(void)
{
  log(L_INFO "Scanning directories");
  scan_empty_dirs("host/");
  scan_empty_dirs("obj/");
}

/*** MAIN ***/

static void help(void) NONRET;

static void
help(void)
{
  fprintf(stderr, "Usage: expire <options>\n\n\
Options:\n\
\n\
-v\t\tBe verbose\n\
-x <file>\tUse extra filter defined in <file>\n\
-f <file>\tDelete URL's listed in <file> only\n\
-t\t\tTest mode -- dump URL's to expire only\n\
-d\t\tForce cleanup of directory structure\n\
-q\t\tRebuild the queue\n\
-Q\t\tRebuild the queue only, no expiration\n\
");
  exit(1);
}

static char options[] = "vx:f:tdqQ";

int
main(int argc, char **argv)
{
  int i;
  char *extra_filter_name = NULL;
  char *custom_list = NULL;
  int test_only = 0;
  int do_expire = 1;

  initlog(argv[0]);
  cf_read("cf/expire", cfitems);
  open_log_file(log_file);
  db_init();
  url_filter_init();

  while ((i = getopt(argc, argv, options)) != -1)
    switch (i)
      {
      case 'v':
	verbose++;
	break;
      case 'x':
	extra_filter_name = optarg;
	break;
      case 'f':
	custom_list = optarg;
	break;
      case 't':
	test_only = 1;
	break;
      case 'd':
	rem_empty = 1;
	break;
      case 'Q':
	do_expire = 0;
	/* Fall-thru */
      case 'q':
	rebuild_queue = 1;
	break;
      default:
	help();
      }
  if (optind != argc)
    help();

  if (test_only && rebuild_queue)
    die("Rebuilding of queue not supported in test mode");
  if (extra_filter_name)
    extra_filter = filter_load(extra_filter_name);

  log(L_INFO "Expire started");

  if (custom_list)
    F = bopen(custom_list, O_RDONLY, 65536);
  else if (do_expire)
    F = bopen("tmp/expire.tmp", O_CREAT | O_TRUNC | O_RDWR, 65536);

  gather_lock();
  urldb_open();
  md5db_open();
  if (rebuild_queue)
    queue_rebuild_start();
  else
    queue_init();

  if (do_expire)
    {
      if (!custom_list)
	{
	  expire_fill(1);
	  bflush(F);
	  bsetpos(F, 0);
	  if (!test_only)
	    {
	      expire_read();
	      bclose(F);
	      unlink("tmp/expire.tmp");
	    }
	}
      else if (!test_only)
	expire_read();
    }
  else
    expire_fill(0);

  if (rem_empty)
    rem_empty_dirs();

  queue_cleanup();
  urldb_close();
  md5db_close();
  gather_unlock();
  log(L_INFO "Expire done: seen %d objects, requeued %d, expired %d, filtered out %d, ignored %d.",
      obj_count, requeue_count, expire_count, filter_count, ignore_count);
  log(L_INFO "Removed %d empty dirs.", empty_dir_count);
  return 0;
}
