/*
 *	Sherlock Gatherer -- Queue Management
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>

#include "gather.h"

#include <lib/url.h>
#include <lib/fastbuf.h>
#include <lib/pools.h>
#include <lib/pagecache.h>

/*** DEBUGGING ***/

#undef QUEUE_DEBUG

#ifdef QUEUE_DEBUG
#define DBG(x...) log(L_DEBUG x)
#else
#define DBG(x...)
#endif

/*** HOST QUEUE ***/

list host_queue;			/* Main host queue (only those with some requests) */
static list active_hosts;		/* Hosts being processed (and dequeued) */
list idle_hosts;			/* Hosts with no queued requests */
static list waiting_hosts;		/* Hosts waiting to be serviced, sorted by wakeup time */
static struct qhost *host_hash[HOST_HASH_SIZE];
static struct qhost *free_list;		/* Fictive host holding list of free fragments */
static struct fastbuf *host_file;
static struct mempool *host_pool;

static inline uns			/* Calculate hash value of given string */
hash(byte *name)
{
  uns h = strlen(name);

  while (*name)
    h = h*37 + *name++;
  return h & (HOST_HASH_SIZE - 1);
}

struct qhost *				/* Remove first host from the host queue */
dequeue_host(uns *waiting, u32 tim)
{
  struct qhost *h = HEAD(host_queue);
  struct qhost *w = HEAD(waiting_hosts);

  *waiting = 0;

  if (w->n.next && w->wait_until <= tim) /* Wait timed out => dequeue waiting host */
    {
      DBG("Dequeued waiting host: <%s>", w->name);
      h = w;
    }
  else if (h->n.next)			/* Process normal queue */
    {
      DBG("Dequeued host: <%s>", h->name);
    }
  else
    {
      if (w->n.next)			/* Just wait? */
	{
	  DBG("Waiting until %d", w->wait_until);
	  *waiting = w->wait_until - tim;
	}
      return NULL;
    }

  rem_node(&h->n);
  add_tail(&active_hosts, &h->n);
  h->status = QHF_ACTIVE;
  return h;
}

struct qhost *				/* Find (possibly queued) host by its signature */
find_host(uns proto, byte *name, uns port)
{
  struct qhost *h;

  for(h = host_hash[hash(name)];
      h && (h->protocol != proto || h->port != port || strcmp(h->name, name));
      h = h->hash)
    ;
  return h;
}

static struct qhost *			/* Create new host entry */
new_host_entry(uns proto, byte *name, uns port, oid_t id)
{
  struct qhost *h = pool_alloc(host_pool, sizeof(struct qhost) + strlen(name));
  uns hh = hash(name);

  bzero(h, sizeof(struct qhost));
  h->hash = host_hash[hh];
  host_hash[hh] = h;
  h->host_id = id;
  h->port = port;
  h->protocol = proto;
  h->status = QHF_NEW;
  strcpy(h->name, name);
  return h;
}

struct qhost *				/* Create new host and make it active */
new_host(uns proto, byte *name, uns port, oid_t id)
{
  struct qhost *h = new_host_entry(proto, name, port, id);

  add_tail(&active_hosts, &h->n);	/* In active list, but marked QHF_NEW */
  return h;
}

static void				/* Write host entry at current position of host file */
write_host_entry(struct fastbuf *f, struct qhost *h)
{
  bputc(f, h->protocol);
  bputw(f, h->port);
  bputl(f, h->host_id);
  h->hf_pos = btell(f);
  bputl(f, h->qf_pos);
  bputl(f, h->qf_last);
  bputl(f, h->obj_count);
  bputs(f, h->name);
  bputc(f, '\n');
}

static void				/* Rewrite host entry with new data */
flush_host_entry(struct qhost *h)
{
  DBG("Flushing <%s> %x %x", h->name, h->qf_pos, h->qf_last);
  bseek(host_file, h->hf_pos, SEEK_SET);
  bputl(host_file, h->qf_pos);
  bputl(host_file, h->qf_last);
  bputl(host_file, h->obj_count);
  if (h->status == QHF_ACTIVE_MODIFIED)
    h->status = QHF_ACTIVE;
}

static inline void			/* Schedule host entry for re-writing */
touch_host_entry(struct qhost *h)
{
  DBG("Touching <%s>", h->name);
  if (h->status == QHF_ACTIVE || h->status == QHF_ACTIVE_MODIFIED)
    h->status = QHF_ACTIVE_MODIFIED;
  else if (h->status != QHF_NEW)
    flush_host_entry(h);
}

static inline void			/* Write free list if it's been modified */
flush_free_list(void)
{
  if (free_list->status == QHF_ACTIVE_MODIFIED)
    flush_host_entry(free_list);
}

static void				/* Flush all modified host entries */
flush_host_cache(void)
{
  struct qhost *h;

  DO_FOR_ALL(h, active_hosts)
    if (h->status == QHF_ACTIVE_MODIFIED)
      flush_host_entry(h);
  flush_free_list();
}

void					/* Add active host at the end of the host queue */
enqueue_host(struct qhost *h, uns force)
{
  DBG("enqueue_host(%s) (status %d)", h->name, h->status);
  flush_free_list();
  if (!force && h->status != QHF_NEW && h->status != QHF_EMPTY)
    return;
  switch (h->status)
    {
    case QHF_NEW:			/* New record -> append it */
      bseek(host_file, 0, SEEK_END);
      write_host_entry(host_file, h);
      rem_node(&h->n);			/* Was in the active queue... */
      break;
    case QHF_ACTIVE_MODIFIED:		/* Modified -> flush it */
      flush_host_entry(h);
      /* Fall-thru */
    case QHF_ACTIVE:
    case QHF_EMPTY:
      rem_node(&h->n);
      break;
    case QHF_QUEUED:
    case QHF_WAITING:
      return;
    case QHF_LOADED:
      break;
    default:
      die("enqueue_host: invalid status=%d", h->status);
    }

  if (h->qf_pos)
    {
      h->status = QHF_QUEUED;
      add_tail(&host_queue, &h->n);
    }
  else					/* No queued items => move to idle queue */
    {
      h->status = QHF_EMPTY;
      add_tail(&idle_hosts, &h->n);
      DBG("%s is idle...", h->name);
    }
  h->wait_until = 0;
}

void					/* Insert active record to the wait queue */
wait_host(struct qhost *h, uns until)
{
  struct qhost *r, *n;

  DBG("%s waits until %d", h->name, until);
  if (h->status == QHF_ACTIVE_MODIFIED)
    flush_host_entry(h);
  if (h->status != QHF_ACTIVE)
    die("wait_host: bad status");
  rem_node(&h->n);
  h->status = QHF_WAITING;
  h->wait_until = until;
  for(r = (void *) &waiting_hosts.head;
      (n = (void *) r->n.next) && n->n.next && n->wait_until < until;
      r = n);
    ;
  insert_node(&h->n, &r->n);
}

/* ITEM QUEUE */

int queue_fd;				/* File handle for the queue file */

static struct page_cache *queue_cache;
static uns queue_size;
uns queue_cache_size = 16;

static uns				/* Return pointer to next page */
qbuf_get_next(struct page *p)
{
  byte *d = p->data;
  return (d[0] << 24) | (d[1] << 16) | (d[2] << 8) | d[3];
}

static void
qbuf_set_next(struct page *p, uns next)	/* Set pointer to next page */
{
  byte *d = p->data;
  d[0] = next >> 24;
  d[1] = next >> 16;
  d[2] = next >> 8;
  d[3] = next;
}

static struct page *			/* Allocate new page in the queue and lock it */
qblock_alloc(void)
{
  struct page *b;
  uns pos;

  if (pos = free_list->qf_pos)
    {
      pos &= ~QUEUE_PAGE_MASK;
      b = pgc_read(queue_cache, queue_fd, pos);
      free_list->qf_pos = qbuf_get_next(b);
      touch_host_entry(free_list);
      DBG("Got block %x from free list", pos);
    }
  else
    {
      pos = queue_size;
      queue_size += QUEUE_PAGE_SIZE;
      b = pgc_get(queue_cache, queue_fd, pos);
      bzero(b->data, QUEUE_PAGE_SIZE);
      DBG("Got block %x from end of file", pos);
    }
  return b;
}

static void				/* Return page to the page pool */
qblock_free(struct page *b)
{
  DBG("Returning block %x to the pool", (int) b->pos);
  bzero(b->data, QUEUE_PAGE_SIZE);
  qbuf_set_next(b, free_list->qf_pos);
  free_list->qf_pos = b->pos + 4;	/* Beware of page 0 */
  touch_host_entry(free_list);
  pgc_mark_dirty(queue_cache, b);
}

struct qitem *				/* Get one item without dequeueing it first */
peek_item(struct qhost *h)
{
  uns pos;
  struct page *b;
  byte *d;
  static struct qitem *qi = NULL;

  if (!qi)
    qi = xmalloc(sizeof(struct qitem) + MAX_URL_SIZE - 1);

  DBG("peek_item(%s)", h->name);
  while (pos = h->qf_pos)
    {
      DBG("...peeking at %x", pos);
      b = pgc_read(queue_cache, queue_fd, pos & ~QUEUE_PAGE_MASK);
      pos &= QUEUE_PAGE_MASK;
      if (b->data[pos])
	{
	  d = qi->text;
	  while (*d = b->data[pos])
	    d++, pos++;
	  qi->aux = b->pos + pos + 1;
	  pgc_put(queue_cache, b);
	  return qi;
	}
      pos = qbuf_get_next(b);
      h->qf_pos = pos;
      qblock_free(b);
      pgc_put(queue_cache, b);
      if (!pos)
	h->qf_last = 0;
      touch_host_entry(h);
    }
  return NULL;
}

struct qitem *				/* Fetch item and dequeue it */
dequeue_item(struct qhost *h)
{
  struct qitem *i = peek_item(h);

  if (i)
    {
      h->qf_pos = i->aux;
      touch_host_entry(h);
    }
  return i;
}

void					/* Add new item to the queue */
enqueue_item(struct qhost *h, byte *name)
{
  struct page *b;
  uns last;
  uns len = strlen(name) + 1;

  DBG("enqueue_item(%s,%s)", h->name, name);
  if (!h->qf_last)
    {
      b = qblock_alloc();
      last = 4;
      DBG("...allocated first block at %x", (int) b->pos);
    }
  else
    {
      b = pgc_read(queue_cache, queue_fd, h->qf_last & ~QUEUE_PAGE_MASK);
      last = h->qf_last & QUEUE_PAGE_MASK;
      DBG("...block at %x, last=%x", (int) b->pos, last);
      if (last + len >= QUEUE_PAGE_SIZE)
	{
	  struct page *o = b;
	  b = qblock_alloc();
	  qbuf_set_next(o, b->pos + 4);
	  pgc_mark_dirty(queue_cache, o);
	  pgc_put(queue_cache, o);
	  last = 4;
	  DBG("...overflow, new block at %x, last=%x", (int) b->pos, last);
	}
    }
  strcpy(b->data + last, name);
  b->data[last+len-1] = 0;
  if (!h->qf_pos)
    h->qf_pos = b->pos + last;
  h->qf_last = b->pos + last + len;
  pgc_mark_dirty(queue_cache, b);
  pgc_put(queue_cache, b);
  touch_host_entry(h);
}

/* INITIALIZATION AND CLEANUP */

static void				/* Load list of all hosts from the host file */
load_host_list(void)
{
  struct fastbuf *f;

  DBG("Loading host list");
  host_file = f = bopen(HOST_FILE_NAME, O_RDWR | O_CREAT, 4096);
  if (bgetc(f) == EOF)
    {
      log(L_INFO "Host database doesn't exist, creating new one");
      free_list = new_host_entry(0, "", 0, 0);
      write_host_entry(f, free_list);
    }
  else
    {
      bsetpos(f, 0);
      for(;;)
	{
	  int proto;
	  uns port, hpos, qpos, lpos, cnt;
	  oid_t id;
	  struct qhost *h;
	  byte name[MAX_URL_SIZE];
		  
	  proto = bgetc(f);
	  if (proto == EOF)
	    break;
	  port = bgetw(f);
	  id = bgetl(f);
	  hpos = btell(f);
	  qpos = bgetl(f);
	  lpos = bgetl(f);
	  cnt = bgetl(f);
	  bgets(f, name, MAX_URL_SIZE);
	  h = new_host_entry(proto, name, port, id);
	  h->qf_pos = qpos;
	  h->qf_last = lpos;
	  h->obj_count = cnt;
	  h->hf_pos = hpos;
	  h->status = QHF_LOADED;
	  if (!h->port && !h->protocol)
	    free_list = h;
	  else
	    enqueue_host(h, 1);
	}
    }
  if (!free_list)
    die("Malformed host file: missing free list entry");
}

static void
flush_host_list(struct fastbuf *f, list *l)
{
  struct qhost *h;

  DO_FOR_ALL(h, *l)
    write_host_entry(f, h);
}

static void				/* Dump all known hosts to the host file */
flush_host_db(void)
{
  struct fastbuf *f;

  DBG("Flushing host database");
  bclose(host_file);
  unlink(HOST_BACK_NAME);
  if (rename(HOST_FILE_NAME, HOST_BACK_NAME) < 0)
    die("Cannot rename " HOST_FILE_NAME " to " HOST_BACK_NAME ": %m");
  f = bopen(HOST_FILE_NAME, O_RDWR|O_CREAT|O_TRUNC, 4096);
  if (!f)
    die("Cannot write host file: %m");
  write_host_entry(f, free_list);
  flush_host_list(f, &active_hosts);
  flush_host_list(f, &host_queue);
  flush_host_list(f, &waiting_hosts);
  flush_host_list(f, &idle_hosts);
  bclose(f);
  unlink(HOST_BACK_NAME);
}

void					/* Initialize the queueing mechanism */
queue_init(void)
{
  DBG("queue_init");
  host_pool = new_pool(16384);
  init_list(&host_queue);
  init_list(&active_hosts);
  init_list(&idle_hosts);
  init_list(&waiting_hosts);
  load_host_list();
  free_list->status = QHF_ACTIVE;
  if ((queue_fd = open(QUEUE_FILE_NAME, O_RDWR | O_CREAT, 0666)) < 0)
    die("Unable to open queue file: %m");
  queue_cache = pgc_open(QUEUE_PAGE_SIZE, queue_cache_size);
  queue_size = lseek(queue_fd, 0, SEEK_END);
}

void					/* Cleanup of the queue machinery */
queue_cleanup(void)
{
  DBG("queue_cleanup");
  flush_host_db();
  pgc_close(queue_cache);
}

void					/* Write out all modified records */
queue_sync(void)
{
  DBG("queue_sync");
  pgc_flush(queue_cache);
#if 0
  flush_host_cache();
#endif
  bflush(host_file);
}
