/*
 *	Sherlock Database Builder -- Object Downloader
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include <lib/lib.h>
#include <lib/lists.h>
#include <lib/searchdb.h>
#include <lib/string.h>
#include <lib/wordcmp.h>
#include <lib/fastbuf.h>
#include <lib/pools.h>
#include <lib/url.h>
#include "filter/filter.h"

/*** PROTOTYPES ***/

static char *find_line(uns c);

struct line {
  struct line *next;
  byte x[1];
};

static struct line *first_line, **last_line;
static struct filter *global_filter;

/*** CONFIGURATION ***/

static byte *field_list = NULL;
static uns rx_timeout = 1000, cache_size = 4096, allowed_word_cats = ~0;
static uns detect_dups;

struct host {
  node n;
  int port;				/* -1 == local */
  byte name[1];				/* Host or directory name */
};

#define WEED_HASH 256

struct weed {
  struct weed *next;
  int len;
  byte w[1];
};

struct weed *weed_hash[WEED_HASH];

static list hosts;

static byte *
cf_host(struct cfitem *c, byte *b)
{
  byte *w[2];
  struct host *h;

  if (wordsplit(b, w, 2) != 2)
    return "Host line syntax error";
  h = xmalloc(sizeof(struct host) + strlen(w[0]));
  strcpy(h->name, w[0]);
  if (sscanf(w[1], "%d", &h->port) != 1 || !h->port)
    return "Invalid port number";
  add_tail(&hosts, &h->n);
  return NULL;
}

static byte *
cf_local(struct cfitem *c, byte *b)
{
  struct host *h;

  h = xmalloc(sizeof(struct host) + strlen(b));
  strcpy(h->name, b);
  h->port = -1;
  add_tail(&hosts, &h->n);
  return NULL;
}

static uns
weed_hash_val(byte *b)
{
  uns h = 0;

  while (*b)
    h = (h*13) + *b++;
  return h & (WEED_HASH - 1);
}

static byte *
cf_weed(struct cfitem *c, byte *b)
{
  while (*b)
    if (Cspace(*b))
      b++;
    else
      {
	byte *c = b;
	uns l, h;
	struct weed *w;

	while (*b && !Cspace(*b))
	  {
	    *b = Cupcase(*b);
	    b++;
	  }
	l = b - c;
	w = xmalloc(sizeof(struct weed) + l);
	w->len = l;
	memcpy(w->w, c, l);
	w->w[l] = 0;
	h = weed_hash_val(w->w);
	w->next = weed_hash[h];
	weed_hash[h] = w;
      }
  return NULL;
}

static struct cfitem cfitems[] = {
  { "Host", CI_FUNCTION, cf_host },
  { "Local", CI_FUNCTION, cf_local },
  { "Fields", CI_STRING, &field_list },
  { "Timeout", CI_INT, &rx_timeout },
  { "CacheSize", CI_INT, &cache_size },
  { "WordCats", CI_INT, &allowed_word_cats },
  { "DetectDups", CI_INT, &detect_dups },
  { "Weed", CI_FUNCTION, cf_weed },
  { NULL, CI_STOP, NULL }
};

/*** VARS ***/

static int sock;
static struct fastbuf *Sock;
static struct fastbuf *obj_file, *obj_idx, *word_ref, *nword_ref;
static uns total_objects;
static u32 section_count[256];

/*** CONNECTIONS ***/

static void
connect_local(struct host *h)
{
  int sox[2];
  int pid;

  log(L_INFO "Downloading from local directory %s", h->name);
  if (chdir(h->name))
    die("chdir() failed: %m");
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, sox))
    die("socketpair() failed: %m");
  pid = fork();
  if (pid < 0)
    die("fork(): %m");
  if (!pid)
    {
      close(0);
      close(1);
      close(2);
      dup(sox[1]);
      dup(sox[1]);
      dup(sox[1]);
      close(sox[0]);
      close(sox[1]);
      execl("bin/gived", "gived", NULL);
      die("exec() failed: %m");
    }
  close(sox[1]);
  sock = sox[0];
}

static void
connect_net(struct host *h)
{
  struct hostent *a;
  struct sockaddr_in si;

  log(L_INFO "Connecting to %s:%d via TCP", h->name, h->port);
  if (! (a = gethostbyname(h->name)))
    die("%s: Unable to resolve: %m", h->name);
  if (a->h_length != 4)
    die("%s: Invalid address length", h->name);
  si.sin_family = AF_INET;
  si.sin_port = htons(h->port);
  memcpy(&si.sin_addr.s_addr, a->h_addr, 4);
  if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_IP)) < 0)
    die("socket: %m");
  if (connect(sock, (struct sockaddr *) &si, sizeof(si)) < 0)
    die("connect(%s): %m", h->name);
}

static void
disconnect_local(void)
{
  int st;
  wait(&st);
  if (st)
    die("Give-daemon failed, status=%d", st);
}

static void
timeout(int unused)
{
  die("Timed out");
}

static void
set_timeout(void)
{
  signal(SIGALRM, timeout);
  alarm(rx_timeout);
}

static byte *
getline(void)
{
  static byte lb[2048];

  if (bgets(Sock, lb, sizeof(lb)))
    return lb;
  die("Unexpected EOF during network transfer");
}

/*** OBJECT MGMT ***/

static struct sdbitem obj;
static byte *server;

static inline void
put_obj_ptr(void)
{
  bputp(obj_idx, btell(obj_file));
}

static void
magic_weight(struct filter_args *a)
{
  /*
   *  This function is called as a part of the document weighing process
   *  and can modify the weight assigned in any way it wants. Really magic
   *  things can happen here. Object attributes can be retrieved via calls
   *  to find_line().
   */
}

static char *
objget_find_attr(struct filter_args *a, int attr)
{
  return find_line(attr);
}

static int
build_obj(uns cnt)
{
  byte *override;
  struct line *l;
  struct filter_args a;
  int weight;

  obj.oid = cnt;
  bzero(&a, sizeof(a));
  a.filter = global_filter;
  a.source = server;
  a.get_attr = objget_find_attr;

  if (!(a.url = find_line('U')))
    a.url="unknown:";

  /* Consult filters */
  if (!filter_run(&a))
    return 0;

  /* Overrides from object itself */
  if (override = find_line('t'))
    a.tag = atol(override);
  if (override = find_line('w'))
    a.weight = atol(override);

  /* Apply magic weighing transforms */
  magic_weight(&a);

  /* Store results in an appropriate way */
  obj.tag = a.tag;
  weight = a.weight;	/* Convert to signed */
#ifdef SHERLOCK_CONFIG_REF_WEIGHTS
  weight += WW_OFFSET;
  obj.weight = (weight < 0) ? 0 : (weight > WW_MAX) ? WW_MAX : weight;
  obj.flags = WQ_OFFSET;
#else
  weight = weight / 100 + WQ_OFFSET;
  obj.flags = (weight < 0) ? 0 : (weight > WQ_MASK) ? WQ_MASK : weight;
#endif

  /* Account for the object */
  section_count[obj.tag]++;

  return 1;
}

/*** WORD LIST ***/

struct cword {
  struct cword *next;
  struct sdbitem ref;
  word len;
  byte w[1];
};

static byte *word_cache, *wc_ptr, *wc_end;
static struct cword *first_cword;

static void
reset_word_cache(void)
{
  first_cword = NULL;
  wc_ptr = word_cache;
}

static int
wc_cmp(struct cword *A, struct cword *B)
{
  uns ta = WORDTYPE(A->ref.flags);
  uns tb = WORDTYPE(B->ref.flags);
  int res;

  if (!ta && tb)
    return -1;
  if (ta && !tb)
    return 1;
  res = ta ? strcmp(A->w, B->w) : search_wordcmp(A->w, B->w);
  if (!res)
    {
      if (A->ref.oid < B->ref.oid)
	return -1;
      if (A->ref.oid > B->ref.oid)
	return 1;
      if (A->ref.flags < B->ref.flags)
	return -1;
      if (A->ref.flags > B->ref.flags)
	return 1;
    }
  return res;
}

static struct cword *
sort_word_list(struct cword *x)
{
  struct cword *l, **ll, *r, **rr, *p;

  if (!x || !x->next)			/* Trivial cases */
    return x;

  ll = &l;				/* Split */
  rr = &r;
  while (x)
    {
      *ll = x;
      ll = &x->next;
      x = x->next;
      if (!x)
	break;
      *rr = x;
      rr = &x->next;
      x = x->next;
    }
  *ll = NULL;
  *rr = NULL;

  l = sort_word_list(l);		/* Sort */
  r = sort_word_list(r);

  ll = &x;				/* Merge */
  while (l && r)
    {
      if (wc_cmp(l, r) < 0)
	{
	  p = l;
	  l = l->next;
	}
      else
	{
	  p = r;
	  r = r->next;
	}
      *ll = p;
      ll = &p->next;
    }
  if (l)
    *ll = l;
  else
    *ll = r;
  return x;
}

static void
flush_word_list(void)
{
  struct cword *cw, *tw;
  u32 n;
  struct fastbuf *f;
  struct sdbitem old;

  if (!first_cword)
    return;
  cw = sort_word_list(first_cword);
  while (cw)
    {
      n = 1;
      old = cw->ref;
      tw = cw;
      cw = cw->next;
      while (   cw
	     && cw->len == tw->len
	     && !!WORDTYPE(cw->ref.flags) == !!WORDTYPE(tw->ref.flags)
	     && !strcmp(cw->w, tw->w))
	{
	  if (memcmp(&cw->ref, &old, sizeof(old)))
	    {
	      n++;
	      old = cw->ref;
	    }
	  else if (detect_dups)
	    log(L_WARN "Duplicate reference for obj %08x, word `%s', zone %02x, flags %02x",
		old.oid, cw->w, cw->ref.tag, cw->ref.flags);
	  cw = cw->next;
	}
      f = WORDTYPE(tw->ref.flags) ? nword_ref : word_ref;
      bwrite(f, tw->w, tw->len+1);
      bputl(f, n);
      old.oid = ~0;
      while (tw != cw)
	{
	  if (memcmp(&tw->ref, &old, sizeof(old)))
	    {
	      old = tw->ref;
	      bwrite(f, &tw->ref, sizeof(tw->ref));
	    }
	  tw = tw->next;
	}
    }
}

static inline void
do_add_word(byte *w, uns l, uns flags)
{
  struct cword *wc;
  uns x;
  struct weed *weed;

  if (l >= MAX_WORD_LEN)
    die("Word too long (%d)", l);

  x = (sizeof(struct cword) + l + CPU_STRUCT_ALIGN - 1) & ~(CPU_STRUCT_ALIGN - 1); /* Magic */
  if (wc_ptr + x > wc_end)
    {
      flush_word_list();
      reset_word_cache();
    }
  wc = (struct cword *) wc_ptr;
  wc->ref = obj;
  wc->ref.flags = flags;
  memcpy(wc->w, w, l);
  wc->w[l] = 0;
  wc->len = l;

  for(weed = weed_hash[weed_hash_val(wc->w)]; weed; weed=weed->next)
    if ((uns) weed->len == l && !strcmp(weed->w, wc->w))
      return;

  wc_ptr += x;
  wc->next = first_cword;
  first_cword = wc;
}

static void
do_word_list(byte *b, int cat)
{
  int weight;
  byte *c;

  if (! (allowed_word_cats & (1 << cat)))
    return;
  b++;
  while (*b)
    {
      if (Cblank(*b))
	{
	  b++;
	  continue;
	}
      if (*b >= 0x80 && *b <= 0x9f)	/* Weight... */
	weight = *b++ - 0x80;
      else
	weight = WQ_OFFSET;
      weight += (obj.flags & WQ_MASK) - WQ_OFFSET;
      if (weight < 0)
	weight = 0;
      if (weight > WQ_MASK)
	weight = WQ_MASK;
      c = b;
      while (*b && !Cblank(*b))
	b++;
      do_add_word(c, b-c, weight | (cat << WT_SHIFT));
    }
}

static void
init_word_cache(void)
{
  word_cache = xmalloc(cache_size);
  wc_end = word_cache + cache_size;
  reset_word_cache();
}

/*** PROCESS SINGLE LINE ***/

static struct mempool *line_pool;

static void
init_line_list(void)
{
  line_pool = new_pool(16384);
}

static inline void
lines_clear(void)
{
  flush_pool(line_pool);
  first_line = NULL;
  last_line = &first_line;
}

static inline void
do_line(byte *l)
{
  struct line *a = fast_alloc(line_pool, sizeof(struct line) + strlen(l));
  strcpy(a->x, l);
  *last_line = a;
  last_line = &a->next;
  a->next = NULL;
}

static char *
find_line(uns c)
{
  struct line *a;

  for(a=first_line; a; a=a->next)
    if (a->x[0] == c)
      return a->x + 1;
  return NULL;
}

static void
go_lines(void)
{
  struct line *l = first_line;

  while (l)
    {
      switch (l->x[0])
	{
	case 'W':
	  do_word_list(l->x, WT_WORD);
	  break;
	case 'R':					/* Normal reference */
	case 'f':					/* Reference to form */
	case 'j':					/* Applet */
	case 'J':					/* Applet */
	  do_word_list(l->x, WT_REF);
	  break;
	case 'I':
	  do_word_list(l->x, WT_REF_IMG);
	  break;
	case 'U':
	  do_word_list(l->x, WT_URL);
	  break;
	}
      if (l->x[1] && strchr(field_list, l->x[0]))
	bputsn(obj_file, l->x);
      l = l->next;
    }
}

/*** HOST AND DOMAIN ***/

static void
do_host_words(struct cword *c, int cat)
{
  if (! (allowed_word_cats & (1 << cat)))
    return;
  c = sort_word_list(c);
  while (c)
    {
      if (!c->next || strcmp(c->next->w, c->w))
	do_add_word(c->w, c->len, (obj.flags & WQ_MASK) | (cat << WT_SHIFT));
      c = c->next;
    }
}

static void
do_host(void)
{
  struct cword *hosts = NULL;
  struct cword *domains = NULL;
  struct cword *w;
  struct line *l;
  int len;

  for(l=first_line; l; l=l->next)
    if (l->x[0] == 'U')
      {
	struct url u;
	byte buf[MAX_URL_SIZE];
	char *url = l->x+1;
	char *c;
	if (url_split(url, &u, buf))
	  {
	    log(L_WARN "Error splitting URL %s", url);
	    continue;
	  }
	if (!u.host)
	  {
	    log(L_WARN "Host-less URL %s", url);
	    continue;
	  }
	len = strlen(u.host);
	w = alloca(sizeof(struct cword) + len);
	w->next = hosts;
	hosts = w;
	w->ref = obj;
	w->len = len;
	strcpy(w->w, u.host);
	c = strchr(u.host, '.');
	if (!c)
	  continue;
	c++;
	len = strlen(c);
	w = alloca(sizeof(struct cword) + len);
	w->next = domains;
	domains = w;
	w->ref = obj;
	w->len = len;
	strcpy(w->w, c);
      }

  do_host_words(hosts, WT_HOST);
  do_host_words(domains, WT_DOMAIN);
}

/*** MAIN DOWNLOADER LOOP ***/

static void
download(struct host *h)
{
  byte *l;
  int obj_count = 0;
  oid_t oid;
  int mod;
  byte xb[16];

  if (h->port < 0)
    connect_local(h);
  else
    connect_net(h);
  Sock = bfdopen(sock, 8192);

  set_timeout();
  l = getline();
  if (l[0] != '+')
    die("%s: Got `%s'", h->name, l);
  write(sock, "0\n", 2);
  server = h->name;

  for(;;)
    {
      set_timeout();
      while (!*(l = getline()))
	;
      if (l[0] != '+')
	die("%s: Got `%s'", server, l);
      if (l[1] == ' ')
	break;
      if (sscanf(l+1, "%8x%d", &oid, &mod) != 2)
	die("%s: Malformed record: %s", server, l);
      lines_clear();
      sprintf(xb, "O%08x", oid);
      do_line(xb);
      while (*(l = getline()))
	do_line(l);
      if (build_obj(total_objects))
	{
	  put_obj_ptr();
	  go_lines();
	  do_host();
	  bputc(obj_file, 0);
	  total_objects++;
	}
      obj_count++;
    }
  if (h->port < 0)
    disconnect_local();
  bclose(Sock);
  log(L_INFO "%s: got %d objects", server, obj_count);
}

int
main(int argc, char **argv)
{
  struct host *h;

  initlog(argv[0]);
  if (argc < 5 || argc > 6)
    die("Usage: objget <obj-file> <obj-idx> <word-ref> <nonword-ref> [<config>]");
  init_list(&hosts);
  cf_read((argc > 5) ? argv[5] : "cf/objget", cfitems);
  global_filter = filter_load_default();
  obj_file = bopen(argv[1], O_WRONLY | O_CREAT | O_TRUNC, 65536);
  obj_idx = bopen(argv[2], O_WRONLY | O_CREAT | O_TRUNC, 65536);
  word_ref = bopen(argv[3], O_WRONLY | O_CREAT | O_TRUNC, 65536);
  nword_ref = bopen(argv[4], O_WRONLY | O_CREAT | O_TRUNC, 65536);
  init_word_cache();
  init_line_list();

  DO_FOR_ALL(h, hosts)
    download(h);
  flush_word_list();
  put_obj_ptr();

  bwrite(obj_file, section_count, sizeof(section_count));
  bclose(obj_file);
  bclose(obj_idx);
  bclose(word_ref);
  bclose(nword_ref);
  log(L_INFO "Total score: %d objects", total_objects);
  return 0;
}
