/*
 *	Sherlock Search Engine -- Configuration File
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>

#include "sherlock.h"

#include <lib/string.h>

static byte *log_file;

uns port = 4444, listen_queue = 4, log_incoming, log_rejected, log_requests;
uns num_matches = 100, cache_size = 10, max_cache_age = 0, max_words = HARD_MAX_WORDS;
uns max_word_matches = HARD_MAX_WORD_MATCHES, file_cache_size = 16, found_word_bonus;
uns default_word_weight = 100, log_timing, connection_timeout = 60;
uns global_debug = 0, global_accent_mode = ACCENT_AUTO, send_all_words;
uns max_wildcard_zone = ~0, max_output_matches = ~0, allow_approx, min_wildcard_prefix_len;
uns ref_weight_scale = 10;
byte *control_password;

struct database *databases;

static list access_list;

struct access_entry {
  node n;
  uns allow;
  u32 addr, mask;
};

static byte *
parse_ip(byte *x, u32 *a)
{
  uns i, q;
  u32 z = 0;

  for(i=0; i<4; i++)
    {
      q = 0;
      while (Cdigit(*x))
	{
	  q = q*10 + *x++ - '0';
	  if (q > 255)
	    return "Invalid IP address";
	}
      if (*x++ != ((i == 3) ? 0 : '.'))
	return "Invalid IP address";
      z = (z << 8) | q;
    }
  *a = z;
  return NULL;
}

static byte *
cf_access(struct cfitem *c, byte *b)
{
  byte *p = strchr(b, '/');
  byte *q;
  struct access_entry *a = xmalloc(sizeof(struct access_entry));

  a->allow = (c->name[0] == 'A');
  if (p)
    {
      *p++ = 0;
      if (q = parse_ip(p, &a->mask))
	return q;
    }
  else
    a->mask = ~0;
  add_tail(&access_list, &a->n);
  return parse_ip(b, &a->addr);
}

int
check_access(u32 ip)
{
  struct access_entry *a;

  DO_FOR_ALL(a, access_list)
    if (! ((ip ^ a->addr) & a->mask))
      return a->allow;
  return 0;
}

static byte *
cf_database(struct cfitem *a, byte *c)
{
  struct database *d = xmalloc(sizeof(struct database));
  byte *fields[3];

  if (wordsplit(c, fields, 3) != 3)
    return "Invalid database definition";
  bzero(d, sizeof(*d));
  d->next = databases;
  databases = d;
  d->obj_pool_name = stralloc(fields[0]);
  d->obj_idx_name = stralloc(fields[1]);
  d->ref_list_name = stralloc(fields[2]);
  return NULL;
}

static struct index *
new_index(int size)
{
  struct index *i = xmalloc(size);

  bzero(i, size);
  if (!databases)
    return NULL;
  i->db = databases;
  i->next = databases->indices;
  databases->indices = i;
  return i;
}

static byte *
cf_words(struct cfitem *a, byte *c)
{
  struct word_index *i;
  byte *fields[2];

  i = (void *) new_index(sizeof(struct word_index));
  if (!i)
    return "No database defined";
  if (wordsplit(c, fields, 2) != 2)
    return "Invalid word index definition";
  i->tree_name = stralloc(fields[0]);
  i->list_name = stralloc(fields[1]);
  i->idx.init = word_init;
  return NULL;
}

static byte *
cf_nonwords(struct cfitem *a, byte *c)
{
  struct nonword_index *i;
  byte *fields[2];

  i = (void *) new_index(sizeof(struct nonword_index));
  if (!i)
    return "No database defined";
  if (wordsplit(c, fields, 2) != 2)
    return "Invalid nonword index definition";
  i->list_name = stralloc(fields[0]);
  i->hash_name = stralloc(fields[1]);
  i->idx.init = nonword_init;
  return NULL;
}

static struct cfitem cfitems[] = {
  { "LogFile", CI_STRING, &log_file },
  { "Database", CI_FUNCTION, cf_database },
  { "Words", CI_FUNCTION, cf_words },
  { "NonWords", CI_FUNCTION, cf_nonwords },
  { "Port", CI_INT, &port },
  { "ListenQueue", CI_INT, &listen_queue },
  { "Allow", CI_FUNCTION, cf_access },
  { "Deny", CI_FUNCTION, cf_access },
  { "LogIncoming", CI_INT, &log_incoming },
  { "LogRejected", CI_INT, &log_rejected },
  { "LogRequests", CI_INT, &log_requests },
  { "NumMatches", CI_INT, &num_matches },
  { "CacheSize", CI_INT, &cache_size },
  { "CacheAge", CI_INT, &max_cache_age },
  { "MaxWords", CI_INT, &max_words },
  { "MaxWordMatches", CI_INT, &max_word_matches },
  { "FileCachePages", CI_INT, &file_cache_size },
  { "WordBonus", CI_INT, &found_word_bonus },
  { "WordWeight", CI_INT, &default_word_weight },
  { "LogTiming", CI_INT, &log_timing },
  { "ConnTimeOut", CI_INT, &connection_timeout },
  { "Debug", CI_INT, &global_debug },
  { "ControlPassword", CI_STRING, &control_password },
  { "AccentMode", CI_INT, &global_accent_mode },
  { "MaxWildcardZone", CI_INT, &max_wildcard_zone },
  { "MinWildcardPrefix", CI_INT, &min_wildcard_prefix_len },
  { "AllowApprox", CI_INT, &allow_approx },
  { "SendAllWords", CI_INT, &send_all_words },
  { "MaxOutputObjects", CI_INT, &max_output_matches },
  { "RefWeightScale", CI_INT, &ref_weight_scale },
  { NULL, CI_STOP, NULL }
};

void
read_config(void)
{
  init_list(&access_list);
  cf_read("cf/sherlockd", cfitems);
  open_log_file(log_file);
}
