/*
 *	Sherlock Search Engine -- Query Processing
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "sherlock.h"

#include <lib/pools.h>
#include <lib/lfs.h>

/* Generation of normalized query strings for easy cache lookup and debugging output */

static byte *
norm_expr(byte *b, struct expr *x)
{
  switch (x->type)
    {
    case EX_AND:
    case EX_OR:
      *b++ = '(';
      b = norm_expr(b, x->u.op.l);
      *b++ = (x->type == EX_AND) ? '&' : '|';
      b = norm_expr(b, x->u.op.r);
      *b++ = ')';
      break;
    case EX_NOT:
      *b++ = '!';
      b = norm_expr(b, x->u.op.l);
      break;
    case EX_MATCH:
      b += sprintf(b, "%d/%04x/%d:\"%s\"/%d", x->u.match.quality,
		   x->u.match.classmap, x->u.match.is_nonword, x->u.match.word, x->u.match.accent_mode);
      break;
    case EX_ANY:
      b += sprintf(b, "ANY");
      break;
    default:
      die("Confused by gazing at unknown type %d", x->type);
    }
  return b;
}

void
normalize_query(struct query *q, byte *buf)
{
  int i;

  for(i=0; i<256/8; i++)
    buf += sprintf(buf, "%02x", q->sections[i]);
  *buf++ = ' ';
  *norm_expr(buf, q->expr) = 0;
}

/* Query cache */

struct cache_item {
  node h;				/* In hash queue (_must_ be first!) */
  node n;				/* In LRU queue */
  byte *request;			/* Normalized request */
  oid_t *oids;				/* Object ID's of resulting objects */
  u32 *qs;				/* Quality factors of resulting objects */
  uns len;				/* Number of objects found */
  time_t timestamp;			/* Timestamp of last successful fill */
  struct mempool *pool;			/* Pool for cached replies */
  struct replybuf rbuf;			/* Cached replies */
};

static uns hash_count;
static list cache_lru, cache_free, *cache_hash;

static int process_query(struct query *, struct cache_item *);
static void show_objects(struct query *, struct cache_item *, uns, uns);

void
cache_init(void)
{
  uns i;

  hash_count = nextprime(cache_size);
  log(L_INFO "Allocating cache for %d items hashed to %d boxes", cache_size, hash_count);
  init_list(&cache_lru);
  init_list(&cache_free);
  cache_hash = xmalloc(sizeof(list) * hash_count);
  for(i=0; i<hash_count; i++)
    init_list(&cache_hash[i]);
  for(i=0; i<cache_size; i++)
    {
      struct cache_item *a = xmalloc(sizeof(struct cache_item));
      a->oids = xmalloc(sizeof(oid_t) * num_matches);
      a->qs = xmalloc(sizeof(u32) * num_matches);
      add_tail(&cache_free, &a->n);
    }
}

static uns
get_hash_val(byte *x)
{
  uns a = strlen(x);

  while (*x)
    a = 23*a + *x++;
  return a % hash_count;
}

static struct cache_item *
get_free_item(void)
{
  struct cache_item *c = SKIP_BACK(struct cache_item, n, HEAD(cache_free)); /* Ugly trick */

  if (c->n.next)
    rem_node(&c->n);
  else
    {
      c = SKIP_BACK(struct cache_item, n, HEAD(cache_lru)); /* :-( */
      if (!c->n.next)
	die("Out of cache slots. Ouch!");
      rem_node(&c->n);
      rem_node(&c->h);
      free_pool(c->pool);
    }
  return c;
}

static struct cache_item *
get_cache_item(struct query *q, uns now)
{
  byte buf[8192];
  struct cache_item *c;
  uns hash;

  normalize_query(q, buf);
  DB(CACHE, "Normalized: %s", buf);
  hash = get_hash_val(buf);
  DO_FOR_ALL(c, cache_hash[hash])
    if (!strcmp(c->request, buf) && (now - c->timestamp) < max_cache_age && !DFLAG(OVERRIDE_CACHE))
      {
	rem_node(&c->n);
	add_tail(&cache_lru, &c->n);
	DB(CACHE, "Found in cache");
	return c;
      }
  c = get_free_item();
  c->timestamp = 0;
  c->pool = new_pool(1024);
  c->request = pool_alloc(c->pool, strlen(buf) + 1);
  strcpy(c->request, buf);
  add_tail(&cache_lru, &c->n);
  add_tail(&cache_hash[hash], &c->h);
  init_reply_buf(&c->rbuf, c->pool);
  return c;
}

void
do_query(struct query *q)
{
  uns now = time(NULL);
  struct cache_item *c = get_cache_item(q, now);
  uns start, count;

  q->parse_msec = get_timer();
  cache_rbuf = &c->rbuf;
  if (!c->timestamp)
    {
      if (!process_query(q, c))
	return;				/* Error */
      c->timestamp = now;
      add_cr("S%d", c->len);
    }
  else
    add_qr("C%d", now - c->timestamp);

  prepend_reply(query_rbuf, "+OK");
  start = q->from;
  if (start > c->len)
    start = c->len;
  count = q->to;
  if (count < start)
    count = start;
  if (count >= c->len)
    count = c->len - 1;
  count = count - start + 1;
  if (count > max_output_matches)
    count = max_output_matches;
  add_qr("N%d", count);
  flush_reply_buf(q, query_rbuf);
  ship_reply_buf(q, cache_rbuf);

  reply_string(q, "\n", 1);
  show_objects(q, c, start, count);
  q->objs_msec = get_timer();

  add_qr("t%d\t%d\t%d\t%d\t%d",
	    q->parse_msec, q->map_msec, q->words_msec, q->refs_msec, q->objs_msec);
  if (log_timing)
    log(L_INFO "Timing: parse=%d map=%d words=%d refs=%d[%d] objs=%d[%d] total=%d",
	q->parse_msec, q->map_msec, q->words_msec, q->refs_msec, c->len, q->objs_msec, count,
	q->parse_msec + q->map_msec + q->words_msec + q->refs_msec + q->objs_msec);
}

/* List of known words */

static struct known_word words[HARD_MAX_WORDS];
static uns num_words;

static int
find_words(struct expr *x)
{
  uns i;
  struct known_word *w;

  switch (x->type)
    {
    case EX_AND:
    case EX_OR:
    case EX_NOT:
      if (!find_words(x->u.op.l) || x->u.op.r && !find_words(x->u.op.r))
	return 0;
      break;
    case EX_MATCH:
      for(i=0; i<num_words; i++)
	{
	  w = &words[i];
	  if (w->is_nonword == x->u.match.is_nonword &&
	      !strcmp(w->text, x->u.match.word) &&
	      w->accent_mode == x->u.match.accent_mode &&
	      w->classmap == x->u.match.classmap)
	  goto match1;
	}
      if (num_words >= max_words || num_words >= HARD_MAX_WORDS)
	return 0;
      w = &words[num_words++];
      strcpy(w->text, x->u.match.word);
      w->is_nonword = x->u.match.is_nonword;
      w->accent_mode = x->u.match.accent_mode;
      w->classmap = x->u.match.classmap;
      w->qfactor = 0;
      w->index = i;
      w->mask = 1 << i;
      w->nrefs_accum = w->approx_flag = 0;
    match1:
      w->qfactor += x->u.match.quality;
      x->u.match.wd_index = w->index;
      DBG("`%s' -> %d (acc=%d)", x->u.match.word, w->index, w->accent_mode);
      break;
    case EX_ANY:
      break;
    default:
      die("What do you mean by showing extype=%d?", x->type);
    }
  return 1;
}

static int
solve_words(struct database *db, struct query *q)
{
  struct index *ix;
  uns i;
  int j;

  for(i=0; i<num_words; i++)
    {
      struct known_word *w = words + i;

      j = 0;
      for(ix=db->indices; ix; ix=ix->next)
	if (j = ix->lookup(ix, w))
	  break;
      DB(WORDS, "[%s] (type=%d, map=%x) -> %d refs", w->text, w->is_nonword, w->classmap, j);
      if (j < -1000000)
	{
	  switch (j)
	    {
	    case -1000001:
	      prepend_reply(query_rbuf, "-111 Pattern <%s> too long", w->text);
	      break;
	    case -1000002:
	      prepend_reply(query_rbuf, "-113 Maximum wildcard zone size (%d) exceeded for <%s>", max_wildcard_zone, w->text);
	      break;
	    case -1000003:
	      prepend_reply(query_rbuf, "-114 Wildcard prefix too short");
	      break;
	    default:
	      prepend_reply(query_rbuf, "-112 Unknown internal error while processing <%s>", w->text);
	    }
	  return 0;
	}
      else if (j < 0)
	{
	  if (!allow_approx)
	    {
	      prepend_reply(query_rbuf, "-105 Too many words matching <%s> [max is <%d>]", w->text, max_word_matches);
	      return 0;
	    }
	  w->approx_flag = 1;
	  j = -j;
	}
      w->nrefs = j;
      w->nrefs_accum += j;
    }
  return 1;
}

static void
emit_word_info(void)
{
  uns i;

  for(i=0; i<num_words; i++)
    {
      struct known_word *w = &words[i];
      add_cr("W%d\t%s", w->nrefs_accum, w->text);
      if (w->approx_flag)
	add_cr("A%s", w->text);
    }
}

/* Boolean expression evaluation */

static byte bool_map[ 1 << (HARD_MAX_WORDS - 3) ];
static uns bool_bytes;

static void
map_bool(struct expr *x, byte *m)
{
  byte *a, *b;
  u32 *X, *Y, *Z;
  uns c;

  Z = (u32 *) m;
  switch (x->type)
    {
    case EX_AND:
    case EX_OR:
      a = alloca(bool_bytes);
      b = alloca(bool_bytes);
      map_bool(x->u.op.l, a);
      map_bool(x->u.op.r, b);
      X = (u32 *) a;
      Y = (u32 *) b;
      if (x->type == EX_AND)
	for(c=0; c<bool_bytes; c += sizeof(u32))
	  *Z++ = *X++ & *Y++;
      else
	for(c=0; c<bool_bytes; c += sizeof(u32))
	  *Z++ = *X++ | *Y++;
      break;
    case EX_NOT:
      a = alloca(bool_bytes);
      map_bool(x->u.op.l, a);
      X = (u32 *) a;
      for(c=0; c<bool_bytes; c += sizeof(u32))
	*Z++ = ~*X++;
      break;
    case EX_MATCH:
      {
	uns i = x->u.match.wd_index;
	if (i < 5)			/* Fits in u32 */
	  {
	    uns k = 1 << i;
	    uns t;
	    u32 j = 0;
	    for(t=0; t < 32; t++)
	      if (t & k)
		j |= 1 << t;
	    for(c=0; c<bool_bytes; c += sizeof(u32))
	      *Z++ = j;
	  }
	else
	  {
	    uns per = 1 << (i - 5);
	    uns t = 2*per*sizeof(u32);
	    uns i;
	    for(c=0; c<bool_bytes; c += t)
	      {
		for(i=0; i<per; i++)
		  *Z++ = 0;
		for(i=0; i<per; i++)
		  *Z++ = ~0;
	      }
	  }
	break;
      }
    case EX_ANY:
      memset(m, ~0, bool_bytes);
      break;
    default:
      die("Lost in mist and scared of seeing ghost #%d", x->type);
    }
#if 0
  {
    int o;
    for(o=bool_bytes; --o >= 0;)
      printf("%02x", m[o]);
    putchar('\n');
  }
#endif
}

static void
construct_bool(struct expr *x)
{
  if (num_words <= 5)
    bool_bytes = sizeof(u32);
  else
    bool_bytes = 1 << (num_words - 3);
  DBG("Constructing boolean expression map of size %d", bool_bytes);
  map_bool(x, bool_map);
}

/* Shortcut solution testing */

static int				/* <0=false, >0=true, 1=don't know */
shortcut_solution(struct expr *x)
{
  int l, r;

  switch (x->type)
    {
    case EX_AND:
      l = shortcut_solution(x->u.op.l);
      r = shortcut_solution(x->u.op.r);
      if (l > 0 && r > 0)
	return 1;
      if (l < 0 || r < 0)
	return -1;
      return 0;
    case EX_OR:
      l = shortcut_solution(x->u.op.l);
      r = shortcut_solution(x->u.op.r);
      if (l > 0 || r > 0)
	return 1;
      if (l < 0 && r < 0)
	return -1;
      return 0;
    case EX_NOT:
      l = shortcut_solution(x->u.op.l);
      return -l;
    case EX_MATCH:
      return (words[x->u.match.wd_index].nrefs) ? 0 : -1;
    case EX_ANY:
      return 1;
    default:
      die("shortcut_solution: black magic (%d)", x->type);
    }
}

static int
shortcut_sections(struct database *db, struct query *q)
{
  uns i;

  for(i=0; i<256/8; i++)
    if (q->sections[i] & db->section_flags[i])
      return 1;
  return 0;
}

/* Reference processing */

static int ref_fd;

struct ref_ref {
  oid_t oid;
  int q;
};

static struct ref_ref *ref_heap;
static uns ref_heap_cnt, ref_heap_total;

struct ref_src {
  node n;
  sh_off_t pos;
  void *locked_page;
  struct sdbitem *refs;
  uns nrefs, mrefs;
  struct known_word *kw;
  struct database *db;
};

static struct ref_src *ref_sources;
static list ref_src_queue;
static uns ref_cnt;
static byte *section_bitmap;

static void
get_ref_srcs(struct database *db)
{
  uns i, j, l;
  struct ref_src *r = ref_sources;
  u32 *z;

  init_list(&ref_src_queue);
  ref_cnt = 0;
  for(i=0; i<num_words; i++)
    {
      struct known_word *w = &words[i];
      uns nmat = 0;
      for(j=0; j<w->nrefs; j++)
	{
	  add_tail(&ref_src_queue, &r->n);
	  r->pos = w->refs[j];
	  z = (u32 *) read_data(db->ref_fd, r->pos, &l);
	  if (l < sizeof(u32))
	    die("Internal error: Misaligned reference entry!");
	  r->pos += sizeof(u32);
	  r->locked_page = lock_last_page();
	  r->refs = (struct sdbitem *) (z+1);
	  r->nrefs = (FILE_PAGE_SIZE - (r->pos & (FILE_PAGE_SIZE-1))) / sizeof(struct sdbitem);
	  r->mrefs = *z;
	  if (r->nrefs > r->mrefs)
	    r->nrefs = r->mrefs;
	  nmat += r->mrefs;
	  r->mrefs -= r->nrefs;
	  r->kw = w;
	  r->db = db;
	  DBG("REF %d: @%d, %d cached, %d more", ref_cnt, r->pos, r->nrefs, r->mrefs);
	  r++;
	  ref_cnt++;
	}
      add_cr("w%d\t%s", nmat, w->text);
    }
}

static void
clean_ref_srcs(void)
{
  uns i;

  for(i=0; i<ref_cnt; i++)
    if (ref_sources[i].locked_page)
      unlock_page(ref_sources[i].locked_page);
}

static uns
fill_in_more(struct ref_src *r)
{
  if (!r->mrefs)
    {
      rem_node(&r->n);
      return 0;
    }
  unlock_page(r->locked_page);
  r->pos = (r->pos + FILE_PAGE_SIZE) & ~(sh_off_t)(FILE_PAGE_SIZE-1);
  r->refs = (struct sdbitem *) read_page(r->db->ref_fd, r->pos);
  r->locked_page = lock_last_page();
  r->nrefs = FILE_PAGE_SIZE / sizeof(struct sdbitem);
  if (r->nrefs > r->mrefs)
    r->nrefs = r->mrefs;
  r->mrefs -= r->nrefs;
  DDBG("Loaded %d more for channel %d (%d remains)", r->nrefs, r - ref_sources, r->mrefs);
  return 1;
}

static inline oid_t
get_obj_id(void)
{
  struct ref_src *rs, *rn, *rb;
  oid_t best;

again:
  best = ~0;
  rb = NULL;
  for(rs = HEAD(ref_src_queue); rn = (void *) rs->n.next; rs = rn)
    {
      if (!rs->nrefs && !fill_in_more(rs))
	continue;
      if (rs->refs->oid < best)
	{
	  best = rs->refs->oid;
	  rb = rs;
	}
    }
  if (rb)
    {
      uns sec = rb->refs->tag;		/* FIXME: This smells of unaligned accesses */
      if (! (section_bitmap[sec >> 3U] & (1 << (sec & 7))))
	{
	  DDBG("OID %08x: bad sec %d", best, sec);
	  DO_FOR_ALL(rs, ref_src_queue)
	    if (rs->refs->oid == best)
	      {
	        rs->refs++;
	        rs->nrefs--;
	      }
	  goto again;
	}
    }
  DDBG("OID %08x", best);
  return best;
}

static inline int
get_obj_q(oid_t obj)
{
  struct ref_src *rs;
  struct known_word *w;
  uns bitmap = 0;
  uns i;
  int bq, allq, q;
  u32 classmask;
  int ref_weight = 0;

  for(i=0; i<num_words; i++)
    words[i].qtemp = 0;

  DO_FOR_ALL(rs, ref_src_queue)
    {
      if (rs->refs->oid != obj)
	continue;
      classmask = 1 << WORDTYPE(rs->refs->flags);
      q = rs->refs->flags & WQ_MASK;
#ifdef SHERLOCK_CONFIG_REF_WEIGHTS
      /* All references to the same object should have the same weight. */
      ref_weight = rs->refs->weight - WW_OFFSET;
#endif
      w = rs->kw;
      if (w->classmap & classmask)
	{
	  bq = (q - WQ_OFFSET) * w->qfactor + found_word_bonus;
	  bitmap |= rs->kw->mask;
	  if (bq >= rs->kw->qtemp)
	    rs->kw->qtemp = bq;
	}
      rs->refs++;
      rs->nrefs--;
    }
  if (! (bool_map[bitmap >> 3U] & (1 << (bitmap & 7))))
    {
      DDBG("Bitmap %x, rejected", bitmap);
      return -1;
    }

  allq = 0;
  for(i=0; i<num_words; i++)
    allq += words[i].qtemp;
  DDBG("Bitmap %x, q=%d+%d, accepted", bitmap, allq, ref_weight * ref_weight_scale);
  allq += ref_weight * ref_weight_scale;
  return (allq >= 0) ? allq : 0;
}

static inline void
heap_remove(uns n)
{
  struct ref_ref *h = ref_heap - 1;	/* We will use indices starting with 1 */
  uns p, z, zz;
  struct ref_ref t;

  h[1] = h[n];
  p = 1;
  for(;;)
    {
      z = 2*p;
      if (z >= n)
	break;
      if ((zz = z+1) >= n)
	zz = z;
      if (h[z].q < h[p].q || h[zz].q < h[p].q)
	{
	  if (h[z].q > h[zz].q)
	    z = zz;
	  t = h[z]; h[z] = h[p]; h[p] = t;
	  p = z;
	}
      else
	break;
    }
}

static inline void
heap_insert(oid_t oid, uns q, uns n)
{
  struct ref_ref *h = ref_heap - 1;	/* We will use indices starting with 1 */
  uns p;

  n++;					/* Position of the new element */
  h[n].oid = oid;
  h[n].q = q;
  while ((p = n/2) && (h[p].q > h[n].q))
    {
      struct ref_ref t;
      t = h[p]; h[p] = h[n]; h[n] = t;
      n = p;
    }
}

static void
fill_heap(struct database *db)
{
  uns heap_count = ref_heap_cnt;
  oid_t obj;
  int q;
  uns total_objs = 0;

  for(;;)
    {
      obj = get_obj_id();
      if (obj == ~0U)			/* END */
	break;
      q = get_obj_q(obj);
      if (q < 0)
	continue;
      if (heap_count >= num_matches)
	heap_remove(heap_count--);
      heap_insert(obj + db->first_id, q, heap_count++);
      total_objs++;
    }
  ref_heap_total += total_objs;
  ref_heap_cnt = heap_count;
}

static void
heap_init(struct query *q)
{
  ref_heap_cnt = ref_heap_total = 0;
  section_bitmap = q->sections;
  section_bitmap[ANY_TAG >> 3U] |= (1 << (ANY_TAG & 7));
}

static void
add_refs(struct database *db)
{
  get_ref_srcs(db);
  fill_heap(db);
  clean_ref_srcs();
}

static uns
do_refs(oid_t *oids, u32 *qs)
{
  uns i;

  add_cr("T%d", ref_heap_total);
  i = ref_heap_cnt;
  while (i)
    {
      --i;
      oids[i] = ref_heap[0].oid;
      qs[i] = ref_heap[0].q;
      DDBG("Ref %d: %08x q=%d", i, oids[i], qs[i]);
      heap_remove(i+1);
    }
  return ref_heap_cnt;
}

void
db_init(void)
{
  struct database *db;
  struct index *idx;
  struct stat st;

  for(db=databases; db; db=db->next)
    {
      db->ref_fd = open_file(db->ref_list_name);
      if (fstat(db->ref_fd, &st) < 0)
	die("Error stating %s: %m", db->ref_list_name);
      db->last_modified = st.st_mtime;
      if (!db->indices)
	log(L_ERROR "No index defined for database %s", db->ref_list_name);
      for(idx=db->indices; idx; idx=idx->next)
	idx->init(idx);
    }

  ref_heap = xmalloc(sizeof(struct ref_ref) * num_matches);
  ref_sources = xmalloc(sizeof(struct ref_src) * max_words * max_word_matches);
}

/* If a cache lookup fails, we must do the hard work... */

static int
process_query(struct query *q, struct cache_item *c)
{
  struct database *db;

  num_words = 0;
  DBG("Unifying words");
  if (!find_words(q->expr))
    {
      prepend_reply(query_rbuf, "-103 Too many words");
      return 0;
    }
  construct_bool(q->expr);
  if (bool_map[0] & 1)			/* Matched by document with none of given words? */
    {
      prepend_reply(query_rbuf, "-104 Negative queries not supported");
      return 0;
    }
  q->map_msec = get_timer();
  heap_init(q);
  for(db=databases; db; db=db->next)
    {
      DB(QUERY, "Database: %s", db->obj_pool_name);
      DBG("Checking sections");
      if (!shortcut_sections(db, q))
	{
	  DB(QUERY, "Shortcut: sections");
	  continue;
	}
      DBG("Searching for words");
      if (!solve_words(db, q))
	{
	  DB(QUERY, "Error");
	  return 0;
	}
      q->words_msec += get_timer();
      DBG("Checking no solution shortcuts");
      if (shortcut_solution(q->expr) < 0)
	{
	  DB(QUERY, "Shortcut: words");
	  continue;
	}
      DBG("Processing references");
      add_refs(db);
      q->refs_msec += get_timer();
    }
  emit_word_info();
  c->len = do_refs(c->oids, c->qs);
  q->refs_msec += get_timer();
  return 1;
}

/* Dumping of objects */

static void
show_objects(struct query *q, struct cache_item *c, uns start, uns count)
{
  while (count--)
    {
      add_qr("Q%d", c->qs[start]);
      add_qr("#%08x", c->oids[start]);
      show_object(q, c->oids[start]);
      flush_reply_buf(q, query_rbuf);
      reply_string(q, "\n", 1);
      start++;
    }
}
