/*
 *	Sherlock Search Engine -- Declarations
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <lib/lib.h>
#include <lib/lists.h>
#include <lib/searchdb.h>

#ifdef BIG_DEBUG
#define DBG(x, y...) log(L_DEBUG x, ##y)
#else
#define DBG(x, y...) do { } while(0)
#endif

#ifdef DETAIL_DEBUG			/* Define locally in the file being debugged... */
#define DDBG(x, y...) DBG(x, ##y)
#else
#define DDBG(x, y...) do { } while(0)
#endif

/* config.c */

void read_config(void);
int check_access(u32);

extern uns port, listen_queue, log_incoming, log_rejected, log_requests;
extern uns num_matches, cache_size, max_cache_age, max_words, max_word_matches;
extern uns file_cache_size, found_word_bonus, default_word_weight, log_timing;
extern uns connection_timeout, global_debug, global_accent_mode, max_wildcard_zone;
extern uns max_output_matches, allow_approx, send_all_words, min_wildcard_prefix_len;
extern uns ref_weight_scale;
extern byte *control_password;

struct database {
  struct database *next;
  struct index *indices;
  oid_t first_id;
  oid_t num_ids;
  byte *obj_pool_name;
  byte *obj_idx_name;
  byte *ref_list_name;
  int obj_pool_fd;
  int obj_idx_fd;
  int ref_fd;
  time_t last_modified;
  u32 section_count[256];
  byte section_flags[256/8];
};

extern struct database *databases;

struct known_word;

struct index {
  struct index *next;
  struct database *db;
  void (*init)(struct index *);
  int (*lookup)(struct index *, struct known_word *);
};

/* request.c */

#define BUF_SIZE 4096

struct reply {
  struct reply *next;
  uns len;
  byte text[1];
};

struct replybuf {
  struct reply *first, **last;
  struct mempool *pool;
};

struct query {
  node n;
  uns established;			/* Time this connection was established */
  byte sections[256/8];			/* Section tag bitmap */
  int sections_set;			/* Section tag bitmap set */
  uns from, to;				/* Result range selection */
  uns debug;				/* Debug flags */
  struct expr *expr;			/* Query expression (NULL if it's a command) */
  char *cmd;				/* Command */
  struct mempool *pool;			/* Memory pool */
  struct replybuf rbuf;			/* Reply header */
  int fd;
  int default_accent_mode;		/* Default word accent mode */
  byte outbuf[BUF_SIZE];		/* Output buffering */
  byte *obptr, *obend;
  byte inbuf[BUF_SIZE];
  byte *ibptr, *ibend;
  uns parse_msec, map_msec,		/* Timing */
      words_msec, refs_msec, objs_msec;
};

extern struct query *current_query;

struct query *new_query(void);
void free_query(struct query *);
void init_reply_buf(struct replybuf *, struct mempool *);
void add_reply(struct replybuf *, byte *, ...);
void prepend_reply(struct replybuf *, byte *, ...);
void reply_string(struct query *, byte *, uns);
void ship_reply_buf(struct query *, struct replybuf *);
void flush_reply_buf(struct query *, struct replybuf *);
void flush_query(struct query *);

extern struct replybuf *query_rbuf, *cache_rbuf;
void add_qr(byte *, ...);
void add_cr(byte *, ...);

#define DFLAG_CACHE 1
#define DFLAG_QUERY 2
#define DFLAG_WORDS 4
#define DFLAG_OVERRIDE_CACHE 8

#ifdef DEBUG
#define DFLAG(flag) (current_query->debug & DFLAG_##flag)
#define DB(flag, x, y...) if (DFLAG(flag)) { log(L_DEBUG x, ##y); add_reply(query_rbuf, "." x, ##y); }
#else
#define DFLAG(flag) 0
#define DB(flag, x, y...)
#endif

/* lex.c */

void lex_init(byte *);
int yylex(void);

/* parse.y */

struct expr {
  byte type;
  union {
    struct {
      struct expr *l, *r;
    } op;
    struct {
      byte *word;
      word quality;
      word classmap;
      byte is_nonword;
      byte wd_index;
      byte accent_mode;
    } match;
  } u;
};

#define EX_AND 0
#define EX_OR 1
#define EX_NOT 2
#define EX_MATCH 3
#define EX_ANY 4

#define ACCENT_AUTO 0			/* STRICT for accented query/match, STRIP elsewhere */
#define ACCENT_STRIP 1			/* strip all accents before comparing */
#define ACCENT_STRICT 2			/* compare with accents */

void err(byte *) NONRET;
byte *parse_query(byte *);

/* query.c */

#define HARD_MAX_WORDS 16		/* Need memory for 2^this bits */
#define HARD_MAX_WORD_MATCHES 256

struct known_word {
  byte text[MAX_WORD_LEN];
  uns is_nonword;
  uns accent_mode;
  uns nrefs;
  uns nrefs_accum;
  uns approx_flag;
  u32 classmap;
  int qfactor;
  int qtemp;
  uns index;
  u32 mask;
  sh_off_t refs[HARD_MAX_WORD_MATCHES];	/* Pointers to the reference index file */
};

void normalize_query(struct query *, byte *);
void cache_init(void);
void do_query(struct query *);
void db_init(void);

/* words.c */

#define BTREE_MAX_DEPTH 5

struct word_index {
  struct index idx;
  byte *tree_name;
  byte *list_name;
  int tree_fd;
  byte *wlist;
  uns wlist_size;
  struct tree_page tree_root;
  uns root_pointers[KEYS_PER_PAGE+1], root_sizes[KEYS_PER_PAGE+1], depths[KEYS_PER_PAGE+1];
};

void word_init(struct index *);
byte *normalize_word(struct mempool *, byte *, uns);

/* nonwords.c */

struct nonword_index {
  struct index idx;
  byte *list_name;
  byte *hash_name;
  int list_fd;
  int hash_fd;
  uns hash_val;
};

void nonword_init(struct index *);
byte *normalize_nonword(struct mempool *, byte *, uns);

/* files.c */

void file_cache_init(void);
int open_file(byte *);
byte *read_page(int, sh_off_t);
byte *read_data(int, sh_off_t, uns *);
void *lock_last_page(void);
void unlock_page(void *);

/* objects.c */

void objects_init(void);
void show_object(struct query *, oid_t);

/* cmds.c */

void do_command(struct query *);
