/*
 *	Sherlock Search Engine -- Searching for Words
 *
 *	(c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

/*
 *	Think thrice before touching. And then don't.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <alloca.h>

#undef BIG_DEBUG
#undef DETAIL_DEBUG

#include "sherlock.h"

#include <lib/string.h>
#include <lib/wordcmp.h>
#include <lib/pools.h>
#include <lib/wildmatch.h>

/* BTREE */

static void
btree_init(struct word_index *ix)
{
  struct tree_page *root;
  uns keys, skip, i, j, w, k, p, ww, pos, rw, depth;
  uns width[BTREE_MAX_DEPTH+1];

  /* See search/build/wtree.c for an explanation of what's happening here. */

  root = (struct tree_page *) read_data(ix->tree_fd, 0, &i);
  memcpy(&ix->tree_root, root, sizeof(ix->tree_root));
  keys = root->aux;
  skip = keys / (KEYS_PER_PAGE + 1);
  i = 0;
  pos = TREE_PAGE_SIZE;
  for(j=0; j<=KEYS_PER_PAGE; j++)
    {
      w = skip;
      if (i >= keys)
	break;
      if (i + w >= keys)
	w = keys - i;			/* Subtree: <w> keys starting with the <i>-th */
      if (w <= 0)
	break;

      width[0] = 0;			/* Calculate subtree width and depth */
      depth = 0;
      while (width[depth] < w)
	{
	  if (depth >= BTREE_MAX_DEPTH)
	    die("Tree too high to climb on");
	  width[depth+1] = width[depth] * (KEYS_PER_PAGE+1) + KEYS_PER_PAGE;
	  depth++;
	}
      ix->depths[j] = depth;

      p = 0;				/* Total pages per subtree */
      rw = 1;				/* Pages per current subtree level */
      for(k=0; k<depth-1; k++)		/* First depth-1 levels are complete */
	{
	  p += rw;
	  rw *= (KEYS_PER_PAGE + 1);
	}
      /*
       * Each bottom level page contains KPP consecutive keys and we skip
       * exactly one key contained in higher level pages between bottom
       * level pages. Therefore the number of such pages is equal to
       * # of keys / (KPP+1) rounded up.
       */
      p += (w + KEYS_PER_PAGE) / (KEYS_PER_PAGE + 1);
      ix->root_sizes[j] = p;
      ix->root_pointers[j] = pos;
      DDBG("subtree=%d first=%d keys=%d depth=%d pages=%d pkeys=%d pos=%d len=%d",
	  j, i, w, depth, p, p*KEYS_PER_PAGE, pos, p * TREE_PAGE_SIZE);
      pos += p * TREE_PAGE_SIZE;
      i += skip + 1;
    }
  DBG("Total tree file size should be %d", pos);
  if ((uns) lseek(ix->tree_fd, 0, SEEK_END) != pos)
    die("Unexpected inconsistency in the btree: length should be %d, but it isn't!", pos);
}

static int
lookup_bt_key(struct tree_page *p, byte *key)
{
  int l = 0;
  int r = KEYS_PER_PAGE - 1;
  int m = 0;
  int z;

  while (l <= r)
    {
      m = (l + r) / 2;
      DDBG("/%-8.8s/%-8.8s/", p->keys[m].sig, key);
      z = memcmp(key, p->keys[m].sig, 8);
      DDBG("l=%d m=%d r=%d z=%d", l,m,r,z);
      if (!z)
	return -m-1;
      if (z < 0)
	r = m - 1;
      else
	l = m + 1;
    }
  z = memcmp(key, p->keys[m].sig, 8);
  return (z > 0) ? (m + 1) : m;
}

static u32
find_sig(struct word_index *ix, byte *sig, int inexact)
{
  /*
   * inexact==0  =>  only exact matches
   * inexact==1  =>  maximum of keys less or equal sig (we know at least one such key exists)
   */

  struct tree_page *r = &ix->tree_root;
  int i, z;
  uns base, pag, npag, dep;
  u32 inex = 0;

  DDBG("Searching for sig /%.8s/ (inex=%d)", sig, inexact);
  i = z = lookup_bt_key(r, sig);
  if (i < 0)
    goto found;
  if (i)
    inex = r->keys[i-1].ptr;
  base = ix->root_pointers[z];
  npag = ix->root_sizes[z];
  pag = 0;
  dep = 0;
  DDBG("Going to subtree %d, base=%d, npag=%d", z, base, npag);
  while (pag < npag)
    {
      r = (struct tree_page *) read_data(ix->tree_fd, TREE_PAGE_SIZE*pag + base, &i);
      i = lookup_bt_key(r, sig);
      if (i < 0)
	goto found;
      if (i)
	inex = r->keys[i-1].ptr;
      dep++;
      if (dep > ix->depths[z])
	{
	  DDBG("Too deep. [DG]iving up.");
	  goto fell;
	}
      pag = pag * (KEYS_PER_PAGE+1) + i + 1;
      DDBG("Going to page %d", pag);
    }
  DDBG("Fell off the tree (pag=%d, npag=%d)", pag, npag);
fell:
  if (inexact)
    {
      DDBG("Found inexact result: sigpos=%d", inex);
      return inex;
    }
  return ~0;

found:
  i = -i-1;
  DDBG("Found at position %d, sigpos=%d", i, r->keys[i].ptr);
  return r->keys[i].ptr;
}

/* Looking up exact words */

static inline int
memcmp_ll(byte *x, byte *y, uns lx, uns ly)
{
  return (lx == ly) ? memcmp(x, y, lx) : 1;
}

static int
lookup_exact(struct word_index *ix, byte *p, byte *pu, uns accent_mode, sh_off_t *x)
{
  byte signature[8];
  u32 sigpos;
  byte *q, *t, *q0;
  uns cnt = 0;
  uns l, r, f;
  byte buf[MAX_WORD_LEN];
  uns pl = strlen(p);
  uns pul = strlen(pu);
  sh_off_t ref;

  extract_sig(p, signature);
  sigpos = find_sig(ix, signature, 0);
  if (sigpos == ~0U)
    return 0;

  q = ix->wlist + sigpos;
  while (*q)
    {
      q0 = q;
      switch (accent_mode)
	{
	case ACCENT_STRICT:
	  l = strlen(q);
	  r = memcmp_ll(q, p, l, pl);
	  q += l + 1;
	  break;
	case ACCENT_STRIP:
	  t = buf;
	  while (*q)
	    {
	      uns uc;
	      GET_UTF8(q, uc);
	      uc = Uunaccent(uc);
	      PUT_UTF8(t, uc);
	    }
	  r = memcmp_ll(buf, pu, t - buf, pul);
	  q++;
	  break;
	default:			/* ACCENT_AUTO */
	  f = 0;
	  while (*q)
	    {
	      uns uc, uu;
	      GET_UTF8(q, uc);
	      uu = Uunaccent(uc);
	      if (uu != uc)
		f = 1;
	    }
	  if (!f)
	    r = memcmp_ll(q0, pu, q - q0, pul);
	  else
	    r = memcmp_ll(q0, p, q - q0, pl);
	  q++;
	}
      ref = 0;
#ifdef CPU_BIG_ENDIAN
      memcpy((byte *) &ref + sizeof(sh_off_t) - BYTES_PER_FILE_POINTER, q, BYTES_PER_FILE_POINTER);
#else
      memcpy(&ref, q, BYTES_PER_FILE_POINTER);
#endif
      q += BYTES_PER_FILE_POINTER;
      DDBG("/%s/ %08x %d", q0, ref, r);
      if (send_all_words)
	add_cr("M%s", q0);
      if (!r)
	{
	  if (cnt++ >= max_word_matches)
	    {
	      DBG("Too many matches");
	      return -max_word_matches;
	    }
	  *x++ = ref;
	}
    }
  return cnt;
}

/* Looking up words with wildcards */

static uns
lookup_wild_range(struct word_index *ix, byte *p, uns *stop)
{
  byte wre[MAX_WORD_LEN];
  byte sig[8];
  byte *z;
  uns start;

  z = wre;
  while (*p != '*' && *p != '?')	/* We know it contains at least one wildcard */
    *z++ = *p++;
  *z = 0;
  if (!wre[0])				/* Leading wildcard -> need to scan whole wordlist */
    {
      *stop = ix->wlist_size;
      return 0;
    }
  extract_sig(wre, sig);
  start = find_sig(ix, sig, 1);
  z = sig;
  while (*z && z <= sig+7)
    z++;
  do
    (*--z)++;
  while (!*z && z > sig);
  if (!*z)
    *stop = ix->wlist_size;
  else
    *stop = find_sig(ix, sig, 1);
  return start;
}

static inline int
match_l(byte *q, struct wildpatt *w, uns l, uns wl)
{
  if (l < wl)
    return 0;
  else
    return wp_match(w, q);
}

static int
lookup_wild(struct word_index *ix, byte *p, byte *pu, uns accent_mode, sh_off_t *x)
{
  struct mempool *pool = current_query->pool;
  struct wildpatt *pw, *puw;
  uns pwl, puwl;
  uns start, stop, l, r, f;
  byte *q, *qe, *q0, *t;
  uns cnt = 0;
  sh_off_t ref;
  byte buf[MAX_WORD_LEN];

  for(q=p; *q != '*' && *q != '?'; q++)
    ;
  if ((unsigned)(q - p) < min_wildcard_prefix_len)
    return -1000003;

  start = lookup_wild_range(ix, p, &stop);
  DDBG("Wildcard range: %d -- %d", start, stop);
  if (start > stop)
    return 0;
  if (stop - start + 1 > max_wildcard_zone)
    return -1000002;
  q = ix->wlist + start;
  qe = ix->wlist + stop + 1;

  if (accent_mode != ACCENT_STRIP)
    {
      pw = wp_compile(p, pool);
      if (!pw)
	return -1000001;
      pwl = wp_min_size(p);
    }
  else
    {
      pw = NULL;
      pwl = 0;
    }
  if (accent_mode != ACCENT_STRICT)
    {
      puw = wp_compile(pu, pool);
      if (!puw)
	return -1000001;
      puwl = wp_min_size(pu);
    }
  else
    {
      puw = NULL;
      puwl = 0;
    }

  while (q < qe)
    {
      while (*q)
	{
	  q0 = q;
	  switch (accent_mode)
	    {
	    case ACCENT_STRICT:
	      l = strlen(q);
	      r = match_l(q, pw, l, pwl);
	      q += l + 1;
	      break;
	    case ACCENT_STRIP:
	      t = buf;
	      while (*q)
		{
		  uns uc;
		  GET_UTF8(q, uc);
		  uc = Uunaccent(uc);
		  PUT_UTF8(t, uc);
		}
	      *t = 0;
	      r = match_l(buf, puw, t - buf, puwl);
	      q++;
	      break;
	    default:			/* ACCENT_AUTO */
	      f = 0;
	      while (*q)
		{
		  uns uc, uu;
		  GET_UTF8(q, uc);
		  uu = Uunaccent(uc);
		  if (uu != uc)
		    f = 1;
		}
	      if (!f)
		r = match_l(q0, puw, q - q0, puwl);
	      else
		r = match_l(q0, pw, q - q0, pwl);
	      q++;
	    }
	  ref = 0;
#ifdef CPU_BIG_ENDIAN
	  memcpy((byte *) &ref + sizeof(sh_off_t) - BYTES_PER_FILE_POINTER, q, BYTES_PER_FILE_POINTER);
#else
	  memcpy(&ref, q, BYTES_PER_FILE_POINTER);
#endif
	  q += BYTES_PER_FILE_POINTER;
	  DDBG("/%s/ %08x %d", q0, ref, r);
	  if (send_all_words)
	    add_cr("M%s", q0);
	  if (r)
	    {
	      if (cnt++ >= max_word_matches)
		{
		  DBG("Too many matches");
		  return -max_word_matches;
		}
	      *x++ = ref;
	    }
	}
      q++;	/* Go to next signature section */
    }

  return cnt;
}

/*
 *  Main word lookup function
 *
 *  result > 0		number of matching words
 *  result < 0		overflow happened, returning -number of matching words
 *  result < 1000000	various error conditions
 */

static int
word_lookup(struct index *index, struct known_word *kw)
{
  struct word_index *ix = (struct word_index *) index;
  uns wlen;
  byte *patt, *patt_unacc, *p, *p_u;
  uns seen_wild = 0;
  uns seen_acc = 0;
  uns accent_mode = kw->accent_mode;
  byte *w = kw->text;

  if (kw->is_nonword)
    return 0;

  wlen = strlen(w) + 1;
  p = patt = alloca(wlen);
  p_u = patt_unacc = alloca(wlen);
  while (*w)
    if (Cspace(*w))
      {
	DBG("Shortcut: spaces -> not a word");
	return 0;
      }
    else if (*w == '?' || *w == '*')
      {
	*p++ = *w;
	*p_u++ = *w++;
	seen_wild++;
      }
    else
      {
	uns uc, ua;
	GET_UTF8(w, uc);
	PUT_UTF8(p, uc);
	ua = Uunaccent(uc);
	PUT_UTF8(p_u, ua);
	if (uc != ua)
	  seen_acc++;
      }
  *p = *p_u = 0;

  if (accent_mode == ACCENT_AUTO && !seen_acc)
    accent_mode = ACCENT_STRIP;
  if (seen_acc)
    DBG("Unaccented: `%s'", patt_unacc);

  return (seen_wild ? lookup_wild : lookup_exact)(ix, patt, patt_unacc, accent_mode, kw->refs);
}

/* Initialize the word lookup mechanism */

void
word_init(struct index *index)
{
  struct word_index *ix = (struct word_index *) index;

  ix->tree_fd = open_file(ix->tree_name);
  if (!(ix->wlist = mmap_file(ix->list_name, &ix->wlist_size)))
    die("Unable to map `%s': %m", ix->list_name);
  btree_init(ix);
  index->lookup = word_lookup;
}

/* Normalize given word by uppercasing it */

byte *
normalize_word(struct mempool *p, byte *x, uns classmap)
{
  byte buf[MAX_WORD_LEN];
  byte *d = buf;
  uns len;

  if (!(classmap & (1 << WT_WORD)))
    return x;

  while (*x)
    {
      uns uc;
      GET_UTF8(x, uc);
      uc = Utoupper(uc);
      PUT_UTF8(d, uc);
      if (d > buf + MAX_WORD_LEN - 3)
	break;
    }
  len = d - buf;
  x = pool_alloc(p, len+1);
  memcpy(x, buf, len);
  x[len] = 0;
  return x;
}
